"""
Data structures relating to building and using the experiment's test
environment, to support environmental data logging for the ATLAS inner
tracker(ITK) pixels multi-module cycling box.

This script is expected to run on a Raspberry Pi.
"""

import collections
import contextlib
import functools
import itertools
import math
import time

import qwiic_tca9548a

from mmcb import sensors


class TestEnvironment:
    """
    Contains test setup hardware configuration information.

    Functions provided here read the configuration file, build the data
    structures to store it, check the configuration, probe the hardware
    to check if everything stated by the configuration file is present and
    operational, and finally to return readings from all attached sensors.
    """

    __slots__ = {
        '_sensors': (
            'A list containing class instances of all the sensors in the test setup.'
        ),
        '_multiplexer_instance': (
            'dict: instances of qwiic_tca9548a.QwiicTCA9548A() used to '
            'communicate with Qwiic multiplexers.'
        ),
        '_unique_mux_addresses': (
            'The set of unique multiplexer I2C addresses used by a number of '
            'reporting functions called by _check_for_configuration_issues().'
        ),
    }

    def __init__(self, configuration_file):
        """
        Read the test setup from a configuration file and build an internal
        representation of all devices found to exist.
        """

        ######################################################################
        # read and diagnose user configuration
        ######################################################################

        self._read_configuration_file(configuration_file)
        configuration_ok = self._check_for_configuration_issues()

        if configuration_ok:

            ##################################################################
            # initialise sensor hardware and basic functionality tests
            ##################################################################

            self._remove_missing_qwiic_multiplexers()

            # Sort sensors by multiplexer I2C address then mux channel order
            # to avoid having to enable and disable the mux channel every
            # time a sensor is read.
            self._sensors.sort()

            self._check_multiplexer_connected_sensors_present()
            self._check_directly_attached_sensors_present()

    def __str__(self):
        return '\n'.join(str(s) for s in self._sensors)

    ##########################################################################
    # file i/o
    ##########################################################################

    def _read_configuration_file(self, configuration_file):
        """
        Read in a configuration CSV file that describes the test setup.

        Each line of the CSV files defines a single sensor.

        The CSV fields are:

        (1) sensor_type - Mandatory field. See mapping variable below for
        details.(2) sensor_name - Mandatory field.(3) qmux_i2c

        The I2C address of the Qwiic multiplexer the sensor is attached to.
        Leave this field empty if the sensor is connected directly to the I2C
        bus.

        (4) qmux_chan

        The Qwiic multiplexer channel the sensor is attached to. Leave this
        field empty if the sensor is connected directly to the I2C bus.

        (5) i2c_addr

        The I2C address of the sensor.

        This field supports two configurations: (a) Each multiplexer channel
        may host a chain of sensors with I2C addresses unique to the
        multiplexer channel, or (b) a sensor directly connected to the I2C
        bus.

        If omitted, this value will be assumed to be the default I2C address
        of the sensor_type.

        (6) thermocouple_type

        Only applicable to thermocouples. Leave this field empty for other
        sensor types.

        (7) analogue_input

        Only applicable to sensor types read via an ADS1015 12-bit ADC and
        analogue multiplexer.

        Example configuration file:

        hyt221, hyt221_ch0_M1,  0x71, 0
        hyt221, hyt221_ch1_env, 0x71, 1
        hyt221, hyt221_ch2,     0x71, 2
        smc,    smc,            0x71, 4
        ntc,    ntc,            0x71, 5,,,2
        tc,     TCk1_env,       0x71, 6, 0x67, K
        tc,     TCk2_M1_vc,     0x71, 6, 0x66, K
        tc,     TCt1_M4_vc,     0x71, 6, 0x65, T
        tc,     TCk3_M2_vc,     0x71, 7, 0x66, K
        tc,     TCk3_M3_vc,     0x71, 7, 0x65, K

        ----------------------------------------------------------------------
        args
            configuration_file : string
                file that exists
        ----------------------------------------------------------------------
        returns
            sensors : list of class instances
            e.g.
                [Hyt221(), Hyt221(), Hyt221(), SmcZse30A01F(), AtlasNtc(),
                 Mcp9601(), Mcp9601(), Mcp9601(), Mcp9601(), Mcp9601()]
        ----------------------------------------------------------------------
        """
        # sensor_type to class mapping
        mapping = {
            'bme680': sensors.Bme680,
            'hyt221': sensors.Hyt221,
            'nau7802': sensors.Nau7802,
            'ntc': sensors.AtlasNtc,
            'shtc3': sensors.Shtc3,
            'smcpfm': sensors.SmcPfm725SF01F,
            'smczse': sensors.SmcZse30A01F,
            'smczse12': sensors.SmcZse30A01F12,
            'tc': sensors.Mcp9601,
        }

        # used for a crude validity check
        lower_bound, upper_bound = 2, 7

        self._sensors = []

        with open(configuration_file, "r", encoding="utf-8") as csvfile:
            for row in csvfile:
                no_comment = row.split("#")[0]

                row = [field.strip() for field in no_comment.split(",")]
                rowlen = len(row)

                if lower_bound <= rowlen <= upper_bound:
                    # (1) pad the list with None values to ensure successful unpacking
                    # (2) convert empty strings to None
                    parts = (
                        x if x else None
                        for x, _ in itertools.zip_longest(row, range(upper_bound))
                    )

                    sensor_type = next(parts)
                    try:
                        self._sensors.append(mapping[sensor_type](*parts))
                    except KeyError:
                        print(f"unknown sensor type: {sensor_type}")

    ##########################################################################
    # report issues in user-supplied configuration
    ##########################################################################

    def _report_duplicate_sensor_names(self):
        """
        Report any duplicate sensor names. These should be unique to avoid
        data being overwritten.
        """
        retval = True

        sensor_names = (sensor.name for sensor in self._sensors)
        counts = collections.Counter(sensor_names)
        duplicates = [k for k, v in counts.items() if v > 1]
        if duplicates:
            retval = False
            print(f'duplicate sensor names: {", ".join(x for x in duplicates)}')

        return retval

    def _report_mux_i2c_not_in_range(self):
        """
        Report any Qwiic multiplexer I2C address outside the allowable range.

        https://learn.sparkfun.com/tutorials/qwiic-mux-hookup-guide
        """
        retval = True

        mux_address_range = range(0x70, 0x77 + 1)
        bad_mux_addresses = {
            a for a in self._unique_mux_addresses if a not in mux_address_range
        }

        if bad_mux_addresses:
            retval = False
            print(
                'Qwiic multiplexer I2C addresses must be in range '
                f'{hex(min(mux_address_range))}-{hex(max(mux_address_range))}, '
                f'found: {", ".join(hex(x) for x in bad_mux_addresses)}'
            )

        return retval

    def _report_mux_channel_not_in_range(self):
        """
        Report and Qwiic multiplexer channels outside the allowable range.
        """
        retval = True

        channel_range = range(8)

        sensors_attached_to_multiplexers = {
            sensor
            for sensor in self._sensors
            if sensor.multiplexer_i2c_address in self._unique_mux_addresses
        }

        for sensor in sensors_attached_to_multiplexers:
            try:
                bad = sensor.multiplexer_channel not in channel_range
            except TypeError:
                retval = False
                print(
                    f'{sensor.description}, {sensor.name} bad mux channel: '
                    f'{sensor.multiplexer_channel}'
                )
            else:
                if bad:
                    retval = False
                    print(
                        f'{sensor.description}, {sensor.name} bad mux channel: '
                        f'{sensor.multiplexer_channel}'
                    )

        return retval

    def _report_mux_missing_address_or_channel(self):
        """Report if the user did not supply both required parameters."""
        retval = True

        for sensor in self._sensors:
            pair = (sensor.multiplexer_i2c_address, sensor.multiplexer_channel)
            if len([x for x in pair if x is None]) == 1:
                retval = False
                print(
                    f'{sensor.description}: '
                    'mux address and channel are both required'
                )

        return retval

    def _report_mux_conflicting_i2c_addresses_within_channels(self):
        """
        Report I2C address conflicts within multiplexer channels.

        Note that sensors on a mux channel with identical i2c addresses
        are not regarded as duplicates if their analogue input channels are
        different.
        """
        retval = True

        for mux_address in self._unique_mux_addresses:
            sensors_on_this_mux = [
                sensor
                for sensor in self._sensors
                if sensor.multiplexer_i2c_address == mux_address
            ]
            unique_channels_on_this_mux = {
                sensor.multiplexer_channel for sensor in sensors_on_this_mux
            }

            for mux_channel in unique_channels_on_this_mux:
                sensors_on_this_mux_channel = (
                    sensor
                    for sensor in sensors_on_this_mux
                    if sensor.multiplexer_channel == mux_channel
                )

                # worst case (1 mux, fully populated channel):
                #     math.comb(255, 2) (n=32385)
                # typical: less than 8 devices per mux channel (n=28)
                duplicates = {
                    sensor1.sensor_i2c_address
                    for sensor1, sensor2 in itertools.combinations(
                        sensors_on_this_mux_channel, 2
                    )
                    if sensor1 == sensor2
                }

                if duplicates:
                    retval = False
                    print(
                        f'conflicting I2C addresses found on mux {hex(mux_address)} '
                        f'channel {mux_channel}: {", ".join(hex(x) for x in duplicates)}'
                    )
        return retval

    def _report_mux_conflict_with_directly_connected_sensors(self):
        """
        Report any i2c address clashes between multiplexers and directly
        connected sensors.
        """
        retval = True

        directly_attached_sensors = {
            sensor.sensor_i2c_address
            for sensor in self._sensors
            if sensor.multiplexer_i2c_address is None
        }

        duplicates = directly_attached_sensors.intersection(self._unique_mux_addresses)

        if duplicates:
            retval = False
            print(
                'conflicting i2c addresses found between mux '
                'and directly attached sensors: '
                f'{", ".join(hex(x) for x in duplicates)}'
            )

        return retval

    def _report_i2c_bus_directly_connected_conflicting_addresses(self):
        """
        Report identical i2c addresses on sensors directly connected to the
        i2c bus. This check is for sensors ONLY. Address clashes between
        multiplexers and sensors are detected in
        _report_mux_conflict_with_directly_connected_sensors.
        """
        retval = True

        # no need to check both multiplexer_i2c_address and multiplexer_channel
        directly_connected = (
            sensor
            for sensor in self._sensors
            if sensor.multiplexer_i2c_address is None
        )
        duplicates = [
            sensor1
            for sensor1, sensor2 in itertools.combinations(directly_connected, 2)
            if sensor1 == sensor2
        ]

        if duplicates:
            retval = False
            print(
                'conflicting i2c addresses found on directly attached sensors: '
                f'{", ".join(hex(x.sensor_i2c_address) for x in duplicates)}'
            )

        return retval

    def _report_conflicting_i2c_addresses_between_channels_and_directly_connected(self):
        """
        Report conflicting i2c addresses between sensors directly attached to
        the I2C bus and those attached via a multiplexer. The latter will
        conflict when the multiplexer/channel is selected.
        """
        retval = True

        i2c_addresses_via_mux = set()

        for mux_address in self._unique_mux_addresses:
            sensors_on_this_mux = [
                sensor
                for sensor in self._sensors
                if sensor.multiplexer_i2c_address == mux_address
            ]
            unique_channels_on_this_mux = {
                sensor.multiplexer_channel for sensor in sensors_on_this_mux
            }

            for mux_channel in unique_channels_on_this_mux:
                i2c_addresses_on_this_mux_channel = [
                    sensor.sensor_i2c_address
                    for sensor in sensors_on_this_mux
                    if sensor.multiplexer_channel == mux_channel
                ]

                i2c_addresses_via_mux.update(i2c_addresses_on_this_mux_channel)

        i2c_addresses_of_directly_attached = {
            sensor.sensor_i2c_address
            for sensor in self._sensors
            if sensor.multiplexer_i2c_address is None
        }

        duplicates = i2c_addresses_via_mux.intersection(
            i2c_addresses_of_directly_attached
        )

        if duplicates:
            retval = False
            print(
                'I2C address conflict between directly attached and '
                f'multiplexer channel attached sensors: {", ".join(hex(x) for x in duplicates)}'
            )

        return retval

    def _report_analogue_mux_channels_out_of_range(self):
        """
        Report analogue multiplexer channels out of range, and that the sensor
        needs to have an analogue channel specified.
        """

        retval = True
        sensors_with_multiple_inputs = [sensors.Ads1015, sensors.Pcf8591]

        for sensor in self._sensors:
            check = functools.partial(isinstance, sensor)
            if any(map(check, sensors_with_multiple_inputs)):
                if sensor.analogue_input not in sensor.supported_analogue_inputs:
                    retval = False
                    inputs = ', '.join(
                        map(str, sensor.supported_analogue_inputs)
                    )
                    print(
                        f'{sensor.description}: analogue input not in '
                        f'{inputs}: {sensor.analogue_input}'
                    )

        return retval

    def _report_sensor_i2c_addresses_out_of_range(self):
        """Report sensor I2C addresses out of range."""

        retval = True

        for sensor in self._sensors:
            # Suppress TypeError since the I2C address may be None at this
            # point. If so, it will be changed to the default I2C address later.
            with contextlib.suppress(TypeError):
                if not sensor.i2c_min <= sensor.sensor_i2c_address <= sensor.i2c_max:
                    retval = False
                    print(
                        f'{sensor.description}, I2C address out of range '
                        f'{hex(sensor.i2c_min)}-{hex(sensor.i2c_max)}: '
                        f'{hex(sensor.sensor_i2c_address)}'
                    )

        return retval

    def _report_sensor_unsupported_thermocouple_type(self):
        """Report missing or unsupported thermocouple type."""

        retval = True

        for sensor in self._sensors:
            if isinstance(sensor, sensors.Mcp9601):
                if sensor.thermocouple_type not in sensor.supported_thermocouple_types:
                    retval = False
                    print(
                        f'{sensor.description} {sensor.name} '
                        f'unrecognised thermocouple type: {sensor.thermocouple_type}'
                    )

        return retval

    def _check_for_configuration_issues(self):
        """
        Make reasonable efforts to identify problems with user-supplied data
        before any attempt is made to communicate with the test setup
        hardware.

        Report findings to the user and terminate if there are any issues,
        since they cannot be fixed without user intervention.

        ----------------------------------------------------------------------
        args : none
        ----------------------------------------------------------------------
        returns
            self._unique_mux_addresses
                no explicit return
        ----------------------------------------------------------------------
        """
        # this will be used in a number of the report functions called below
        self._unique_mux_addresses = {
            sensor.multiplexer_i2c_address
            for sensor in self._sensors
            if sensor.multiplexer_i2c_address is not None
        }

        reports = (
            self._report_duplicate_sensor_names,
            self._report_mux_i2c_not_in_range,
            self._report_mux_channel_not_in_range,
            self._report_mux_missing_address_or_channel,
            self._report_mux_conflicting_i2c_addresses_within_channels,
            self._report_mux_conflict_with_directly_connected_sensors,
            self._report_i2c_bus_directly_connected_conflicting_addresses,
            self._report_conflicting_i2c_addresses_between_channels_and_directly_connected,
            self._report_analogue_mux_channels_out_of_range,
            self._report_sensor_i2c_addresses_out_of_range,
            self._report_sensor_unsupported_thermocouple_type,
        )
        results = [report() for report in reports]
        configuration_ok = all(results)
        if not configuration_ok:
            self._sensors.clear()
            print(
                'Please rectify issue(s) with the configuration file and/or test setup.'
            )

        return configuration_ok

    ##########################################################################
    # remove configured hardware that cannot be found
    ##########################################################################

    def _remove_missing_qwiic_multiplexers(self):
        """Remove sensors attached to missing Qwiic multiplexers."""
        self._multiplexer_instance = {}

        for mux_address in self._unique_mux_addresses.copy():

            # The constructor for .QwiicTCA9548A() doesn't check the device
            # exists, therefore doesn't raise an exception or display an error
            i2c_mux = qwiic_tca9548a.QwiicTCA9548A(mux_address)

            # .is_connected() prints an error if the mux cannot be found:
            # "Error connecting to Device: 71, [Errno 121] Remote I/O error".
            # It doesn't raise an exception, but does correctly return False.
            if i2c_mux.is_connected():
                self._multiplexer_instance[mux_address] = i2c_mux
            else:
                print(f'qwiic mux not found at I2C address {hex(mux_address)}')
                self._unique_mux_addresses.remove(mux_address)

                for sensor in self._sensors.copy():
                    if sensor.multiplexer_i2c_address == mux_address:
                        self._sensors.remove(sensor)
                        print(f'removing sensor {sensor.name}')

    def _check_multiplexer_connected_sensors_present(self):
        """
        Check if multiplexer-attached devices are responsive.

        ASSUME that sensors have been sorted by mux i2c address then by
        mux channel number. This allows the list of sensors to be traversed
        minimising the number of changes of mux/channel configuration.

        The initial instances required to access the sensors are created
        during the traversal.
        """

        sensors_attached_to_multiplexers = [
            s for s in self._sensors if s.multiplexer_i2c_address
        ]

        address = self._sequence_change(
            s.multiplexer_i2c_address for s in sensors_attached_to_multiplexers
        )
        channel = self._sequence_change(
            s.multiplexer_channel for s in sensors_attached_to_multiplexers
        )

        i2c_mux = None
        for sensor in sensors_attached_to_multiplexers:

            mux_change, _ = next(address)
            chan_change, mux_chan_previous = next(channel)

            if mux_change:
                i2c_mux = self._multiplexer_instance[sensor.multiplexer_i2c_address]
                i2c_mux.disable_all()

            if chan_change:
                if mux_chan_previous is not None and not mux_change:
                    i2c_mux.disable_channels(mux_chan_previous)

                i2c_mux.enable_channels(sensor.multiplexer_channel)
                time.sleep(0.05)

            sensor.activate()
            if sensor.missing():
                self._sensors.remove(sensor)

        # ensure all channels are off on exit
        if i2c_mux is not None:
            i2c_mux.disable_all()

    def _check_directly_attached_sensors_present(self):
        """
        Check if directly attached sensors are responsive.

        The initial instances required to access the sensors are created
        during the traversal.
        """
        directly_attached_sensors = (
            s for s in self._sensors if not s.multiplexer_i2c_address
        )

        for sensor in directly_attached_sensors:
            sensor.activate()

            if sensor.missing():
                self._sensors.remove(sensor)

    ##########################################################################
    # read sensors
    ##########################################################################

    def read_all_sensors(self):
        """
        ----------------------------------------------------------------------
        args : none
        ----------------------------------------------------------------------
        returns
            data_for_current_timestamp: dict
            e.g. {
                'timestamp': 1650734619.065016,
                'smc_vacuum': -0.3395540799999992,
                'ntc_temperature': -61.19908797184263,
                'hyt221_ch0_M1_temperature': 20.26735030214246,
                'hyt221_ch0_M1_relative_humidity': 0.0,
                'hyt221_ch1_env_temperature': 20.54934993590917,
                'hyt221_ch1_env_relative_humidity': 0.09766221082829762,
                'hyt221_ch2_temperature': 20.579564182384175,
                'hyt221_ch2_relative_humidity': 0.0,
                'TCk1_env_temperature': 20.125,
                'TCk2_M1_vc_temperature': 20.1875,
                'TCt1_M4_vc_temperature': 21.0625,
                'TCk3_M2_vc_temperature': 22.0,
                'TCk3_M3_vc_temperature': 21.875
            }
        ----------------------------------------------------------------------
        """
        data_for_current_timestamp = {'timestamp': time.time()}

        sensors_attached_to_multiplexers = [
            s for s in self._sensors if s.multiplexer_i2c_address
        ]

        address = self._sequence_change(
            s.multiplexer_i2c_address for s in sensors_attached_to_multiplexers
        )
        channel = self._sequence_change(
            s.multiplexer_channel for s in sensors_attached_to_multiplexers
        )

        i2c_mux = None
        for sensor in sensors_attached_to_multiplexers:

            mux_change, _ = next(address)
            chan_change, mux_chan_previous = next(channel)

            if mux_change:
                i2c_mux = self._multiplexer_instance[sensor.multiplexer_i2c_address]
                i2c_mux.disable_all()

            if chan_change:
                if mux_chan_previous is not None and not mux_change:
                    i2c_mux.disable_channels(mux_chan_previous)

                i2c_mux.enable_channels(sensor.multiplexer_channel)

            if not sensor.missing():
                time.sleep(0.05)
                with contextlib.suppress(TypeError):
                    data_for_current_timestamp.update(sensor.read())

        # ensure all channels are off on exit
        if i2c_mux is not None:
            i2c_mux.disable_all()

        # read directly connected sensors

        directly_attached_sensors = (
            s for s in self._sensors if not s.multiplexer_i2c_address
        )
        for sensor in directly_attached_sensors:
            time.sleep(0.05)
            with contextlib.suppress(OSError, TypeError):
                data_for_current_timestamp.update(sensor.read())

        if len(data_for_current_timestamp) == 1:
            data_for_current_timestamp.clear()

        return data_for_current_timestamp

    ##########################################################################
    # support functions
    ##########################################################################

    @staticmethod
    def _sequence_change(number_sequence):
        """
        Used for flagging changes in the incoming number series.

        ASSUMING the number sequence is sorted, this may be used - for example -
        to indicate when channel numbers change. At which point a command can be
        issued to make the change on the hardware.

        e.g. given an incoming generator supplying these multiplexer channel
        numbers:

        (0, 1, 2, 2, 2, 3, 3)

        Changes would be indicated thus:

        ((True, ...), (True, ...), (True, ...), (False, ...), (False, ...),
        (True, ...), (False, ...))

        ----------------------------------------------------------------------
        args
            number_sequence : generator
        ----------------------------------------------------------------------
        yields : tuple (bool, int)
            (True, ...) if the current number in the sequence differed from
            its predecessor, (False, ...) otherwise.
        ----------------------------------------------------------------------
        """
        last = None

        for number in number_sequence:
            if number != last:
                yield True, last
                last = number
            else:
                yield False, last

    ##########################################################################
    # dew point
    ##########################################################################

    def calculate_dew_points(self, data_for_current_timestamp):
        """
        Calculate dew point data from a dictionary of sensor values.

        Dew points are only calculated for sensors that supply both relative
        humidity and temperature; there's no mixing of data from different
        sensors.

        Note that the internal _dew_point method which this method relies
        upon, returns values low values clamped to -71 degrees Celsius, which
        is the figure for the dry air supply in the cleanroom. This is to
        cope with some sensors reporting zero RH% in dry conditions due to
        limited resolution, which would in turn lead to dew point figures
        that would be impossible to realise in the test environment.

        ----------------------------------------------------------------------
        args
            data_for_current_timestamp : dict
                e.g.
                {'timestamp': 1651224444.4431129,
                 'hyt221_ch0_M1_temperature': 20.277421717634134,
                 'hyt221_ch0_M1_relative_humidity': 0.0,
                 'hyt221_ch1_env_temperature': 20.60977842885918,
                 'hyt221_ch1_env_relative_humidity': 0.0, ...}
        ----------------------------------------------------------------------
        returns
            dew_point_calculations: dict
                e.g.
                {'bme_ambient_calculated_dew_point': 7.317889854346651,
                 'hyt221_ch2_calculated_dew_point': -71,
                 'hyt221_ch1_env_calculated_dew_point': -71,
                 'hyt221_ch0_M1_calculated_dew_point': -71,
                 'sht_ambient_calculated_dew_point': 8.439206561314366}
        ----------------------------------------------------------------------
        """
        dew_point_calculations = {}

        rhu = '_relative_humidity'
        tdc = '_temperature'
        # find all keys containing relative humidity or temperature
        columns_of_interest = {
            c for c in data_for_current_timestamp if rhu in c or tdc in c
        }

        # find all sensors that collect both relative humidity and temperature
        sensor_rhp = {c.rsplit(rhu)[0] for c in columns_of_interest if rhu in c}
        sensor_dgc = {c.rsplit(tdc)[0] for c in columns_of_interest if tdc in c}
        sensor_dew = sensor_rhp.intersection(sensor_dgc)

        # create new columns with calculated dew points
        for sensor in sensor_dew:
            # find keys for relative humidity and temperature for this sensor
            source_columns = {x for x in columns_of_interest if sensor in x}

            try:
                column_humi = next(x for x in source_columns if rhu in x)
                column_temp = next(x for x in source_columns if tdc in x)
            except StopIteration:
                continue

            # make it clear this is not directly acquired data
            new_column_title = f'{sensor}_calculated_dew_point'

            # calculate dew point, add new key with this information
            temp = data_for_current_timestamp[column_temp]
            humi = data_for_current_timestamp[column_humi]
            dew_point_calculations[new_column_title] = self._dew_point(temp, humi)

        return dew_point_calculations

    @staticmethod
    def _dew_point(tdc, rhp, arden_buck=True):
        """
        Calculate the dew point given temperature and humidity.

        https://en.wikipedia.org/wiki/Dew_point

        Murray, F. W., 1967, On the Computation of Saturation Vapor Pressure

        Magnus, Tetens : pressure in mbar
        t = 20.1 (deg C)
        u = 7.5
        v = 237.3
        w = 0.7858 (vapor pressure in mbar)

        In [25]: math.pow(10, (t * u)/(t + v) + w)
        Out[25]: 23.521463268069194

        For low temperature and low pressure with accuracy, use Goff-Gratch.

        ----------------------------------------------------------------------
        args
            tdc : float
                temperature degrees Celsius
            rhp : float
                relative humidity percentage
            arden_buck : bool
                choice of algorithm
        ----------------------------------------------------------------------
        returns : float
            dew point degrees Celsius
        ----------------------------------------------------------------------
        """
        # constants
        # a = 6.1121 # mbar
        b = 18.678
        c = 257.14  # °C
        d = 234.5  # °C

        try:
            if arden_buck:
                # accurate
                ymtrh = math.log(
                    (rhp / 100) * math.exp((b - tdc / d) * (tdc / (c + tdc)))
                )
                tdp = (c * ymtrh) / (b - ymtrh)
            else:
                # approximation (Magnus formula)
                ytrh = math.log(rhp / 100) + (b * tdc) / (c + tdc)
                tdp = (c * ytrh) / (b - ytrh)
        except ValueError:
            # clamp the dew point values to -71 which is the figure for the dry air
            # supply in the cleanroom. This is to cope with some sensors reporting
            # zero RH% in dry conditions due to limited resolution.
            tdp = -71

        # Sensors returning RH% values very close to zero due to limited
        # resolution can be problematic. So again as above, clamp the dew point
        # values to -71 deg C.
        return max(tdp, -71)



"""
Common functions

callable functions from this file:

    atomic_send_command_read_response
    cache_read
    check_current
    check_file_exists
    check_ports_accessible
    data_read
    data_write
    decimal_quantize
    dew_point
    exclude_channels
    initial_power_supply_check
    interpret_numeric
    iseg_value_to_float
    list_split
    log_with_colour
    missing_ports
    ports_to_channels
    rate_limit
    read_aliases
    read_psu_measured_vi
    report_output_status
    round_safely
    rs232_port_is_valid
    save_plot
    send_command
    set_psu_voltage
    set_psu_voltage_and_read
    si_prefix
    stage_speed
    synchronise_psu
    timestamp_to_utc
    time_axis_adjustment
    unique
    wait_for_voltage_to_stabilise
    write_consignment_csv

data structures:

    ANSIColours
    Channel
    Consignment
    DEVICE_CACHE
    Packet
    UNIT
"""

import argparse
import bz2
import collections
import contextlib
import copy
import datetime
import decimal
import itertools
import json
import logging
import os
import pickle
import re
import sys
import time
import types

import matplotlib
# agg is used only for writing plots to files, not to the window manager
# this option is set to avoid problems running scripts on remote hosts
# over ssh, matplotlib.use must be called in this exact position
matplotlib.use('agg')
import matplotlib.pyplot as plt
import numpy as np
import serial

from mmcb import lexicon


##############################################################################
# data structures
##############################################################################


DEVICE_CACHE = os.path.expanduser('~/.cache.json')


UNIT = types.SimpleNamespace(
    calculated_dew_point='DP\u00b0C',
    flow_rate='lps',
    pressure='hPa',
    relative_humidity='RH%',
    strain='arbitrary',
    temperature='\u00b0C',
    vacuum='kPa',
)


class ANSIColours:
    """
    ANSI 3/4 bit terminal escape sequences

    https://en.wikipedia.org/wiki/ANSI_escape_code
    """
    # foreground colours
    FG_BRIGHT_BLACK = '\033[90m'
    FG_BRIGHT_RED = '\033[91m'
    FG_BRIGHT_GREEN = '\033[92m'
    FG_BRIGHT_YELLOW = '\033[93m'
    FG_BRIGHT_BLUE = '\033[94m'
    FG_BRIGHT_MAGENTA = '\033[95m'
    FG_BRIGHT_CYAN = '\033[96m'
    FG_BRIGHT_WHITE = '\033[97m'

    FG_BLACK = '\033[30m'
    FG_RED = '\033[31m'
    FG_GREEN = '\033[32m'
    FG_YELLOW = '\033[33m'
    FG_BLUE = '\033[34m'
    FG_CYAN = '\033[35m'
    FG_MAGENTA = '\033[36m'
    FG_WHITE = '\033[37m'

    # background colours
    BG_BRIGHT_BLACK = '\033[100m'
    BG_BRIGHT_RED = '\033[101m'
    BG_BRIGHT_GREEN = '\033[102m'
    BG_BRIGHT_YELLOW = '\033[103m'
    BG_BRIGHT_BLUE = '\033[104m'
    BG_BRIGHT_MAGENTA = '\033[105m'
    BG_BRIGHT_CYAN = '\033[106m'
    BG_BRIGHT_WHITE = '\033[107m'

    BG_BLACK = '\033[40m'
    BG_RED = '\033[41m'
    BG_GREEN = '\033[42m'
    BG_YELLOW = '\033[43m'
    BG_BLUE = '\033[44m'
    BG_MAGENTA = '\033[45m'
    BG_CYAN = '\033[46m'
    BG_WHITE = '\033[47m'

    # attributes
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    BLINK = '\033[5m'
    UNDERLINE = '\033[4m'


class Packet:
    """
    This class holds all data acquired from a single power supply channel
    during a single run of iv.py.

    The IV and IT data may represent one of the following three scenarios:

    * A single outbound IV
    * A single outbound IV and IT
    * An outbound IV, IT, and return IV
    """
    __slots__ = {
        'manufacturer': 'The manufacturer of the power supply.',
        'model': 'The model name of the power supply.',
        'serial_number': 'The serial number of the power supply.',
        'channel': 'Which power supply channel was data acquired from.',
        'ident': 'hardware identifier string or user supplied alias',
        'set_voltage': 'IV data: list of set voltages.',
        'measured_current': 'IV data: list of measured leakage currents at each set voltage',
        'measured_voltage': 'IV data: list of measured bias voltages at each set voltage',
        'measured_timestamp': 'IV data: timestamp of each data acquisition',
        'measured_temperature': 'IV data: temperature (°C) at each data acquisition',
        'measured_humidity': 'IV data: humidity (RH%) at each data acquisition',
        'hold_voltage': 'IT data: constant set voltage for IT test',
        'hold_current': 'IT data: list of measured leakage currents at each set voltage',
        'hold_timestamp': 'IT data: timestamp of each data acquisition',
        'hold_temperature': 'IT data: temperature (°C) at each data acquisition',
        'hold_humidity': 'IT data: humidity (RH%) at each data acquisition'}

    def __init__(self, manufacturer, model, serial_number, channel, ident):
        self.manufacturer = manufacturer
        self.model = model
        self.serial_number = serial_number
        self.channel = channel
        self.ident = ident
        self.set_voltage = []
        self.measured_current = []
        self.measured_voltage = []
        self.measured_timestamp = []
        self.measured_temperature = []
        self.measured_humidity = []
        self.hold_voltage = None
        self.hold_current = []
        self.hold_timestamp = []
        self.hold_temperature = []
        self.hold_humidity = []

    def __repr__(self):
        return (f'Packet('
                f'manufacturer="{self.manufacturer}", '
                f'model="{self.model}", '
                f'serial_number="{self.serial_number}", '
                f'channel="{self.channel}", '
                f'ident="{self.ident}", '
                f'set_voltage={self.set_voltage}, '
                f'measured_current={self.measured_current}, '
                f'measured_voltage={self.measured_voltage}, '
                f'measured_timestamp={self.measured_timestamp}, '
                f'measured_temperature={self.measured_temperature}, '
                f'measured_humidity={self.measured_humidity}, '
                f'hold_voltage={self.hold_voltage}, '
                f'hold_current={self.hold_current}, '
                f'hold_timestamp={self.hold_timestamp}, '
                f'hold_temperature={self.hold_temperature}, '
                f'hold_humidity={self.hold_humidity})')

    def __str__(self):
        # verbose power supply identifier
        identifier = f'{self.manufacturer} {self.model}'
        if self.serial_number:
            identifier += f' s.no. {self.serial_number}'
        if self.channel:
            identifier += f' channel {self.channel}'

        # number of data points in IV and IT tests
        try:
            split_index = self.measured_current.index('split')
        except ValueError:
            outbound_iv_len = len(self.measured_current)
            return_iv_len = 0
        else:
            outbound_iv_len = len(self.measured_current[:split_index])
            return_iv_len = len(self.measured_current[split_index + 1:])

        out_iv_text = f'Outbound IV data points : {outbound_iv_len}'
        hold_it_text = f'IT data points : {len(self.hold_current)}'
        ret_iv_text = f'Return IV data points : {return_iv_len}'

        return '\n'.join([identifier, out_iv_text, hold_it_text, ret_iv_text])

    # sufficient to allow a list of class instances to be sorted
    def __lt__(self, value):
        return (f'{self.model}{self.serial_number}{self.channel}'
                < f'{value.model}{value.serial_number}{value.channel}')

    def extract_environmental_data(self, category, test_iv, outbound=True):
        """
        Extract measurements from their originally stored state to one that is
        more readily usable for data analysis.

        For the desired data source, the dict within the list represents
        measurements from all the available sensors within the category at a
        given point in time (for IT tests) or per voltage step (for IV tests).

        e.g.
        [{'PT100MK1-DC3D6': 22.31, 'PT100MK1-DC392': 19.16}, ...]

        These lists may have had a 'split' marker added by function
        run_iv_test() in iv.py if both outbound and return IV tests were
        requested.

        ----------------------------------------------------------------------
        args
            category : string
                'temperature' or 'humidity'
            test_iv : bool
                True = data from IV test, False = data from IT test
            outbound : bool
                For lists that may contain a 'split' marker, outbound = True
                selects items before the marker, False selects items after the
                marker. For other lists, this is ignored.
        ----------------------------------------------------------------------
        returns : dict
            e.g.
            {'PT100MK1-DC3D6': [21.25, 21.25, 21.26, 21.26, ..., 21.24],
             'PT100MK1-DC392': [20.9, 20.92, 20.92, 20.94, ..., 20.94]}
        ----------------------------------------------------------------------
        """
        assert category in {'temperature', 'humidity'}, 'unknown category'

        # select original data source
        temp = category == 'temperature'
        if test_iv:
            var = self.measured_temperature if temp else self.measured_humidity
            out, ret = list_split(var)
            var = out if outbound else ret
        else:
            var = self.hold_temperature if temp else self.hold_humidity

        # transform selected data
        summary = collections.defaultdict(list)

        for measurement in var:
            for sensor, reading in measurement.items():
                summary[sensor].append(reading)

        # return transformed data source
        return dict(summary)

    def extract_timestamp_data(self, test_iv, outbound=True):
        """
        Extract timestamps to match data obtained by
        self.extract_environmental_data.

        ----------------------------------------------------------------------
        args
            test_iv : bool
                True = data from IV test, False = data from IT test
            outbound : bool
                For lists that may contain a 'split' marker, outbound = True
                selects items before the marker, False selects items after the
                marker. For other lists, this is ignored.
        ----------------------------------------------------------------------
        returns : list
            e.g.
                [21.25, 21.25, 21.26, 21.26, ..., 21.24]
        ----------------------------------------------------------------------
        """
        # select original data source
        if test_iv:
            out, ret = list_split(self.measured_timestamp)
            var = out if outbound else ret
        else:
            var = self.hold_timestamp

        return var


class Consignment:
    """
    Contains data for an entire data acquisition session, which may include
    packets from multiple power supplies, each of which may have multiple
    channels.
    """
    __slots__ = {
        'label': 'Plot title string.',
        'safe_label': ('Either a simplified version of the label string that may be\n'
                       'appended to a filename and is easy to work with on the command\n'
                       'line, or None'),
        'aliases': ('A dictionary containing mappings between power supply channel\n'
                    'hardware identifiers and user-submitted descriptions of each\n'
                    'channel\'s purpose.'),
        'forwardbias': ('A boolean indicating whether the user specified forward bias\n'
                        'operation.'),
        'hold': 'A boolean indicating whether the user specified an IT test',
        'environmental_data_present': ('A boolean indicating that environmental sensors'
                                       'are presented.'),
        'packets': ('A list containing the data packets captured during an entire\n'
                    'data acquisition session.')}

    def __init__(self, label, aliases, forwardbias, hold, environmental_data_present):
        self.label = label
        self.aliases = aliases
        self.forwardbias = forwardbias
        self.hold = hold
        self.environmental_data_present = environmental_data_present
        self.packets = []

        # create a safe and easy to work with label for appending to a
        # filename later
        if self.label:
            self.safe_label = ''.join(c
                                      for c in self.label.replace(' ', '-')
                                      if c not in r'<>:"/\|?*')
        else:
            self.safe_label = None

    def __repr__(self):
        return (f'Consignment('
                f'label="{self.label}", '
                f'aliases={self.aliases}, '
                f'forwardbias={self.forwardbias}, '
                f'hold={self.hold}, '
                f'environmental_data_present={self.environmental_data_present})')

    def __str__(self):
        pretty = [f'label: {self.label}', f'safe_label: {self.safe_label}',
                  f'aliases: {self.aliases}', f'forwardbias: {self.forwardbias}',
                  f'hold: {self.hold}',
                  f'environmental_data_present: {self.environmental_data_present}',
                  f'number of packets: {len(self.packets)}']

        return '\n'.join(pretty)

    def remove_bad_packets(self):
        """
        Only retain usable packets. This operation should be performed before
        attempting to plot or store data.
        """
        self.packets = [packet
                        for packet in self.packets
                        if packet is not None and len(packet.set_voltage) > 1]


class Channel:
    """
    Handle power supplies on a per-channel basis.
    """
    __slots__ = {
        'port': 'Serial port on which to communicate with the power supply.',
        'config': 'Serial port configuration parameters.',
        'serial_number': 'The serial number of the power supply.',
        'model': 'The model name of the power supply.',
        'manufacturer': 'The manufacturer of the power supply.',
        'channel': 'Which power supply channel to use.',
        'category': 'Tailor behaviour for high or low voltage use.',
        'release_delay': 'Delay between sequential serial port interactions.',
        'ident': 'hardware identifier string or user supplied alias.',
        'window_size': 'Size of measured_voltages and measured_currents queues.',
        'measured_voltages': 'Queue to store voltages for smoothing.',
        'measured_currents': 'Queue to store currents for smoothing.'}

    def __init__(self, port, config, serial_number, model, manufacturer,
                 channel, category, release_delay, alias):
        self.port = port
        self.config = config
        self.serial_number = serial_number
        self.model = model
        self.manufacturer = manufacturer
        self.channel = channel
        self.category = category
        self.release_delay = release_delay
        self.window_size = 10
        self.measured_voltages = collections.deque([], self.window_size)
        self.measured_currents = collections.deque([], self.window_size)

        _parameters = (self.model, self.serial_number, self.channel)
        _ident_key = '_'.join(p for p in _parameters if p).lower()
        _text = ' '.join(p for p in _parameters if p).lower()
        try:
            _text = alias.get(_ident_key, _text)
        except AttributeError:
            # alias is None
            pass
        finally:
            self.ident = _text

    def __repr__(self):
        return (f'Channel('
                f'port="{self.port}", config={self.config}, '
                f'serial_number="{self.serial_number}", '
                f'model="{self.model}", '
                f'manufacturer="{self.manufacturer}", '
                f'channel="{self.channel}", '
                f'category="{self.category}", '
                f'release_delay={self.release_delay}, '
                f'ident="{self.ident}")')

    def __str__(self):
        description = f'{self.manufacturer} {self.model} s.no. {self.serial_number}'
        description += f' ch. {self.channel}' if self.channel else ''
        return description

    # sufficient to allow a list of class instances to be sorted
    def __lt__(self, value):
        return (f'{self.model}{self.serial_number}{self.channel}'
                < f'{value.model}{value.serial_number}{value.channel}')

    # sufficient to allow a class instance to be removed from a list
    def __eq__(self, value):
        return (f'{self.model}{self.serial_number}{self.channel}'
                == f'{value.model}{value.serial_number}{value.channel}')


##############################################################################
# file i/o
##############################################################################

def _file_is_compressed(filename):
    """
    Determine if a data file is compressed from its extension.

    --------------------------------------------------------------------------
    args
        filename : string
    --------------------------------------------------------------------------
    returns
        restored_progress : boolean
            True if the filename extension indicates a compressed file,
            False otherwise
    --------------------------------------------------------------------------
    """
    return 'bz2' in os.path.splitext(filename)[-1].lower()


def configcache_read(filename):
    """
    Retrieve previously stored serial port configuration and attached devices
    from human-readable cache file.

    --------------------------------------------------------------------------
    args
        filename : string
    --------------------------------------------------------------------------
    returns
        restored_progress : dict
            deserialized cached data or None
    --------------------------------------------------------------------------
    """
    restored_progress = None

    if os.path.isfile(filename):
        with open(filename, 'r', encoding='utf-8') as infile:
            try:
                restored_progress = json.load(infile)
            except json.JSONDecodeError:
                sys.exit(f'exiting: problem reading data from {filename}')

    return restored_progress


def configcache_write(data, filename):
    """
    Store serial port configuration and attached devices to cache in
    human-readable form.

    --------------------------------------------------------------------------
    args
        data : dict
            a record each detected device, its respective category and
            serial port configuration.
        filename : string
    --------------------------------------------------------------------------
    returns : none
    --------------------------------------------------------------------------
    """
    with open(filename, 'w', encoding='utf-8') as outfile:
        try:
            json.dump(data, outfile)
        except TypeError:
            sys.exit(f'exiting: problem writing data to {filename}')


def data_read(filename):
    """
    Retrieve previously stored data from file.

    --------------------------------------------------------------------------
    args
        filename : string
    --------------------------------------------------------------------------
    returns
        restored_progress : unpickled data or None
    --------------------------------------------------------------------------
    """
    restored_progress = None

    if os.path.isfile(filename):
        file_read = bz2.open if _file_is_compressed(filename) else open

        with file_read(filename, 'rb') as infile:
            try:
                restored_progress = pickle.load(infile)
            except (pickle.UnpicklingError, AttributeError, EOFError,
                    ImportError, IndexError, TypeError):
                sys.exit(f'exiting: problem reading data from {filename}')

    return restored_progress


def data_write(data, filename):
    """
    Store data to file for later retrieval.

    --------------------------------------------------------------------------
    args
        data : data structure to store
        filename : string
    --------------------------------------------------------------------------
    returns : none
    --------------------------------------------------------------------------
    """
    file_write = bz2.open if _file_is_compressed(filename) else open

    with file_write(filename, 'wb') as outfile:
        try:
            pickle.dump(data, outfile)
        except pickle.PicklingError:
            sys.exit(f'exiting: problem writing data to {filename}')


def read_aliases(settings, filename):
    """
    Read in alias file, and write results to settings:
        (1) create a dictionary of aliases
        (2) create a dictionary of channels to ignore

    Comments starting with a # are allowed.

    e.g.

    # Keithley 2614b
    ,2614b,4428182,a,layer front # blue pcb v1.0a
    ,2614b,4428182,b,layer rear  # green pcb v1.0b
    # Keithley 2410 (borrowed from cleanroom)
    ,2410,4343654,,neutron detector
    off,2410,1390035,,layer external # off-axis

    which would generate aliases as indicated in the log:

    INFO : alias 2614b_4428182_a -> layer front
    INFO : alias 2614b_4428182_b -> layer rear
    INFO : alias 2410_4343654 -> neutron detector

    --------------------------------------------------------------------------
    args
        settings : dictionary
            contains core information about the test environment
        filename : string
            filename with extension
    --------------------------------------------------------------------------
    returns
        settings : dict
            no explicit return, mutable type amended in place
    --------------------------------------------------------------------------
    """
    if os.path.exists(filename):
        alias = {}
        ignore = {}

        with open(filename, 'r', encoding='utf-8') as csvfile:
            for line_num, row in enumerate(csvfile):
                # remove comments
                no_comment = row.split('#')[0].strip()

                # separate comma separated values
                fields = (field.strip() for field in no_comment.split(','))
                try:
                    enable, model, serialnum, channel, description = fields
                except ValueError:
                    # not enough values to unpack
                    if no_comment:
                        print(f'line {line_num} (expected 5 fields): {no_comment}')
                    continue

                identifier = '_'.join(x for x in [model, serialnum, channel] if x).lower()

                if description:
                    alias[identifier] = description.replace('"', '').replace('\'', '')

                if enable.lower() in {'no', 'off', 'disable'}:
                    ignore[identifier] = True

        settings['alias'] = alias if alias else None
        settings['ignore'] = ignore if ignore else None
    else:
        print(f'file {filename} could not be read from')


def save_plot(settings, filename):
    """
    Save plot to mass storage in the requested file format.

    --------------------------------------------------------------------------
    args
        settings : dictionary
            contains core information about the test environment
        filename : string
            filename without extension
    --------------------------------------------------------------------------
    returns : none
    --------------------------------------------------------------------------
    """
    if settings['svg']:
        plt.savefig(f'{filename}.svg')
    else:
        plt.savefig(f'{filename}.png', dpi=400)


def write_consignment_csv(consignment, row):
    """
    Write recorded data in consignment to mass storage.

    ASSUMES a file has been opened for writing.

    --------------------------------------------------------------------------
    args
        consignment : instance of class Consignment
            contains data for the whole data acquisition session
        row : csv.writer
            writer object responsible for converting data to delimited strings
    --------------------------------------------------------------------------
    returns : none
    --------------------------------------------------------------------------
    """
    # --label: write label if user has supplied it
    if consignment.label is not None:
        row.writerow(itertools.chain(['label'], [consignment.label]))

    # --alias: write aliases if user has supplied them
    if consignment.aliases is not None:
        for channel_identifier, alias in consignment.aliases.items():
            row.writerow(itertools.chain(['alias'], [channel_identifier], [alias]))

    stored_as_dict = {'measured_temperature', 'hold_temperature',
                      'measured_humidity', 'hold_humidity'}
    not_iterable = {'channel', 'hold_voltage', 'ident',
                    'manufacturer', 'model', 'serial_number'}

    # write all data packets
    for packet in consignment.packets:

        # extract data from instance of class Packet in (key, value) pairs
        # if the value contains something
        # use __slots__ to maintain ordering
        packet_data = {k: getattr(packet, k) for k in packet.__slots__
                       if getattr(packet, k)}

        for field, item in packet_data.items():
            if field in not_iterable:
                row.writerow(itertools.chain([field], [item]))
            elif field in stored_as_dict:
                # list will contain dicts that need to be extracted, e.g.
                # [{'PT100MK1-DC3D6': 20.52, 'PT100MK1-DC392': 19.9},
                #  {'PT100MK1-DC3D6': 20.52, 'PT100MK1-DC392': 19.89}, ...]
                tmd = collections.defaultdict(list)
                for temp_measurement in item:
                    if temp_measurement != 'split':
                        for sensor, reading in temp_measurement.items():
                            tmd[sensor].append(reading)
                    else:
                        for sensor in tmd:
                            tmd[sensor].append(temp_measurement)

                # write the rows to the csv file
                for sensor, measurements in tmd.items():
                    row.writerow(itertools.chain([f'{field} ({sensor})'], measurements))
            else:
                # item is a plain list
                row.writerow(itertools.chain([field], item))


##############################################################################
# identifier/alias management
##############################################################################

def _alias_to_log(settings):
    """
    Enter the mapping of power supply channel identifiers to aliases into the
    log.

    --------------------------------------------------------------------------
    args
        settings : dictionary
            contains core information about the test environment
    --------------------------------------------------------------------------
    returns : none
    --------------------------------------------------------------------------
    """
    if settings['alias'] is not None:
        for ident, alias in settings['alias'].items():
            logging.info('alias: %s -> %s', ident, alias)


def _lookup_reference(packet, use_channel=True, separator=' '):
    """
    For the given data packet, return the hardware identifier for its power
    supply channel.

    --------------------------------------------------------------------------
    args
        packet : instance of class Channel
            Contains data for a given power supply channel's IV and IT curves.
        use_channel : bool
            If True, use channel identifier in returned string.
            Useful to disable when working at power supply level
            rather than channel-level, e.g. when checking power supply
            interlocks and creating error logs that would be misleading if
            they contained a channel identifier.
        separator : string
            Separator character used between values, typically ' ' or '_'.
    --------------------------------------------------------------------------
    returns : string
        e.g.
            use_channel=True
                '2410 1272738' (single channel)
                '2614b 4428182 b' (dual channel)
            use_channel=False
                '2410 1272738' (single channel)
                '2614b 4428182' (dual channel)
    --------------------------------------------------------------------------
    """
    text = f'{packet.model}'
    if packet.serial_number:
        text += f'{separator}{packet.serial_number}'
    if packet.channel and use_channel:
        text += f'{separator}{packet.channel}'

    return text


def _lookup_decorative(alias, packet):
    """
    Decorative print of the power supply channel with its alias (if it
    exists).

    --------------------------------------------------------------------------
    args
        alias : dictionary
            maps aliases to power supply channel identifiers
        packet : packet : instance of class Packet or class Channel
            contains data for a given power supply channel's IV and IT curves
    --------------------------------------------------------------------------
    returns : string
        e.g.
        'keithley 2614b s.no. 4428182 channel b (layer rear)'
        or
        'keithley 2614b s.no. 4428182 channel a'
    --------------------------------------------------------------------------
    """
    parameters = (packet.model, packet.serial_number, packet.channel)
    ident_key = '_'.join(p for p in parameters if p).lower()

    dtext = f'{packet.manufacturer} {packet.model}'
    if packet.serial_number:
        dtext += f' s.no. {packet.serial_number}'
    if packet.channel:
        dtext += f' channel {packet.channel}'

    try:
        atext = alias[ident_key]
    except (KeyError, TypeError):
        suffix = ''
    else:
        suffix = f' ({atext})'

    return f'{dtext}{suffix}'


##############################################################################
# utilities
##############################################################################


def decimal_quantize(value, decimal_places):
    """
    Returns a decimal rounded to the user-specified number of decimal places.

    --------------------------------------------------------------------------
    args
        value : numeric or string representation of numeric
        decimal_places : positive integer
    --------------------------------------------------------------------------
    returns : decimal.Decimal
    --------------------------------------------------------------------------
    """
    value = decimal.Decimal(value)

    # 0: '0', 1: '0.1', 2: '0.01' etc...
    decplc = f'{pow(10, -decimal_places):.{decimal_places}f}'

    return value.quantize(decimal.Decimal(decplc), rounding=decimal.ROUND_HALF_UP)


def dew_point(tdc, rhp):
    """
    Calculate the dew point given temperature and humidity.

    https://en.wikipedia.org/wiki/Dew_point

    Murray, F. W., 1967, On the Computation of Saturation Vapor Pressure

    Magnus, Tetens : pressure in mbar
    t = 20.1 (deg C)
    u = 7.5
    v = 237.3
    w = 0.7858 (vapor pressure in mbar)

    In [25]: math.pow(10, (t * u)/(t + v) + w)
    Out[25]: 23.521463268069194

    For low temperature and low pressure with accuracy, use Goff-Gratch.

    --------------------------------------------------------------------------
    args
        tdc : float
            temperature degrees Celsius
        rhp : float
            relative humidity percentage
    --------------------------------------------------------------------------
    returns : float
        dew point degrees Celsius
    --------------------------------------------------------------------------
    """
    # constants
    # a = 6.1121 # mbar
    b = 18.678
    c = 257.14  # °C
    d = 234.5   # °C

    ymtrh = np.log((rhp / 100) * np.exp((b - tdc / d) * (tdc / (c + tdc))))
    tdp = (c * ymtrh) / (b - ymtrh)

    return tdp


def interpret_numeric(val):
    """
    Convert numbers in either engineering or scientific notation from string
    form to a float.

    Useful resource: https://regex101.com

    --------------------------------------------------------------------------
    args
        val : string
    --------------------------------------------------------------------------
    returns
        success : bool
        val : float if conversion was successful, otherwise return the
            problematic string
    --------------------------------------------------------------------------
    """
    success = True

    eng_notation = re.match(r'^[+\-]?\d*\.?\d*[yzafpnumkMGTPEZY]$', val)
    sci_notation = re.match(r'^[+\-]?(?:\d+\.|\d+\.\d+|\.\d+|\d+)(?:[eE][+\-]?\d+)?$', val)

    if eng_notation is not None:
        value = float(eng_notation[0][:-1])
        suffix = eng_notation[0][-1]
        val = value * pow(10, 'yzafpnum kMGTPEZY'.find(suffix) * 3 - 24)
    elif sci_notation is not None:
        val = float(sci_notation[0])
    else:
        success = False

    return success, val


def iseg_value_to_float(value):
    """
    For values read back from ISEG SHQ power supplies.

    While set voltages are specified as positive-only magnitudes (with
    polarity defined by the hardware switch on the rear panel), the values
    read back have the correct polarity.

    e.g.

    Set PSU channel 1 voltage to 120V with:

    D1=120 (sets voltage)
    G1 (requests power supply to ramp up to given voltage)

    poll G1 until it responds with S1=ON

    U1 returns with (polarity, significand, signed exponent):
    -01200-01

    which this function returns as -120.0

    --------------------------------------------------------------------------
    args
        value : string
            response from ISEG command
    --------------------------------------------------------------------------
    returns
        fvalue : float
    --------------------------------------------------------------------------
    """
    fvalue = None
    significand, exponent = value[:-3], value[-3:]

    with contextlib.suppress(ValueError):
        fvalue = float(f'{significand}e{exponent}')

    return fvalue


def list_split(data):
    """
    Split list at the 'split' marker.

    --------------------------------------------------------------------------
    args
        data : list
    --------------------------------------------------------------------------
    returns
        out : list
        ret : list
    --------------------------------------------------------------------------
    """
    try:
        split_index = data.index('split')
    except ValueError:
        out = data
        ret = []
    else:
        out = data[:split_index]
        ret = data[split_index + 1:]

    return out, ret


def round_safely(start, stop, step, first=True):
    """
    Round the value to the given step "in the direction of travel" so the
    resultant value is always contained within the bounds of start and stop.

    --------------------------------------------------------------------------
    args
        start : numeric (int, float or decimal.Decimal)
            start value of the number sequence
        stop : numeric (int, float or decimal.Decimal)
            stop value of the number sequence
        step : int
            step size
        first : bool
            if True, round the start value, if False, round the stop value
    --------------------------------------------------------------------------
    returns : int
        rounded and aligned value
    --------------------------------------------------------------------------
    """
    step = abs(step)
    value, step = (float(start), step) if first else (float(stop), -step)
    return int(value - (value % -step if start < stop else value % step))


def si_prefix(value, dec_places=3, compact=True):
    """
    Provide an approximation of the given number in engineering
    representation, to the given number of decimal places, with the SI
    (Systeme Internationale) unit prefix appended.

    --------------------------------------------------------------------------
    args
        value : string, float or int
            numeric value
        dec_places : int
            number of decimal places to display
        compact : bool
            if True remove trailing zeros from number, if the remainder is a
            whole number, remove the trailing decimal point
            e.g.
            103.100 -> 103.1
             10.000 ->  10
    --------------------------------------------------------------------------
    returns : string or None
        value with SI unit prefix
    --------------------------------------------------------------------------
    """
    # make sure the number is in scientific notation
    # then separate value and exponent
    try:
        significand, exponent = f'{float(value):e}'.lower().split('e')
    except TypeError:
        return None

    significand = float(significand)
    exponent = int(exponent)

    # align with 10**3 boundaries
    while exponent % 3 != 0:
        exponent -= 1
        significand *= 10

    if -24 <= exponent <= 24:
        # derive SI unit prefix
        if exponent == 0:
            prefix = ''
        else:
            prefix = 'yzafpnum kMGTPEZY'[8 + exponent // 3]

        # remove trailing zeroes
        # if the number is a whole number, remove the trailing decimal point as well
        significand = f'{significand:.{dec_places}f}'
        if compact:
            # avoid rstrip('0.') to ensure '0.0' doesn't become ''
            significand = significand.rstrip('0').rstrip('.')
    else:
        # handle the case where the supplied value is too large or too small
        prefix = ''
        significand = float(value)

    return f'{significand}{prefix}'


def timestamp_to_utc(tref):
    """
    Converts a timestamp into a string in UTC to the nearest second.

    e.g. 1567065212.1064236 converts to '20190829_075332'

    --------------------------------------------------------------------------
    args
        tref : float
            time in seconds since the epoch
    --------------------------------------------------------------------------
    returns : string
    --------------------------------------------------------------------------
    """
    utc = datetime.datetime.utcfromtimestamp(tref).isoformat().split('.')[0]
    return utc.replace('-', '').replace(':', '').replace('T', '_')


def time_axis_adjustment(data):
    """
    Generate the parameters necessary to transform absolute UNIX-style epoch
    timestamps to relative timestamps with human readable units.

    --------------------------------------------------------------------------
    args
        data : pandas.DataFrame
    --------------------------------------------------------------------------
    returns
        units : string
        data : pandas.DataFrame
            no explicit return, mutable type amended in place
    --------------------------------------------------------------------------
    """
    earliest_timestamp = data['timestamp'].min()
    latest_timestamp = data['timestamp'].max()

    minutes_of_available_data = (latest_timestamp - earliest_timestamp) / 60

    if minutes_of_available_data > 180:
        units = 'hours'
        scale = 3600
    else:
        units = 'minutes'
        scale = 60

    data['timestamp'] = (data['timestamp'] - earliest_timestamp) / scale

    return units


##############################################################################
# command line argument processing
##############################################################################

def check_current(val):
    """
    Checks current values.

    Note that the Keithley 2410 will not allow compliance values to be set
    to less than 0.1% of the measurement range, p.18-69 (p.449 in PDF).

    --------------------------------------------------------------------------
    args
        val : string
    --------------------------------------------------------------------------
    returns
        val : float
    --------------------------------------------------------------------------
    """
    success, val = interpret_numeric(val)

    if not success:
        raise argparse.ArgumentTypeError(
            f'{val}: '
            'should be a value in engineering or scientific notation')

    if val < 1e-9:
        raise argparse.ArgumentTypeError(
            f'{val}: '
            'is too small for range of the power supply')

    return val


def check_file_exists(filename):
    """
    check if file exists

    --------------------------------------------------------------------------
    args
        val : string
            filename, e.g. '20200612_132725_psuwatch.log'
    --------------------------------------------------------------------------
    returns : string
    --------------------------------------------------------------------------
    """
    if not os.path.exists(filename):
        raise argparse.ArgumentTypeError(f'{filename}: file does not exist')

    return filename


##############################################################################
# cache import
##############################################################################

def _logprint(uselog, level, message):
    """
    Display via logging module or print as appropriate.

    --------------------------------------------------------------------------
    args
        uselog : bool
            use logging if True, print otherwise
        level : int
            logging level e.g. logging.DEBUG
        text : string
            text to display
    --------------------------------------------------------------------------
    returns : none
    --------------------------------------------------------------------------
    """
    if uselog:
        log_with_colour(level, message)
    else:
        print(message)


def cache_read(device_types=None, uselog=False):
    """
    Get the details of all the given devices from the cache, and in the process
    transform the dictionary from having device types as keys, to having
    port names as keys.

    --------------------------------------------------------------------------
    args
        device_types : list or None
            device types to search cache for e.g. ['hvpsu', 'lvpsu']
            valid types are: controller, daq, hvpsu, lvpsu
            if this is None, do not check for validity, import all devices
        uselog : bool
            use print if False (default), otherwise use the logging module
    --------------------------------------------------------------------------
    returns
        found : dict
            {port: ({port_config}, device_type, device_serial_number), ...}
    --------------------------------------------------------------------------
    """
    cache = configcache_read(DEVICE_CACHE)

    if cache is None:
        _logprint(uselog, logging.CRITICAL,
                  'please run detect.py before using this script (cache missing)')
        sys.exit()

    # if the plaform differs from the platform the cache was created on, the
    # serial port data contained in the cache will almost certainly be wrong
    cached_platform = cache.pop('platform', None)
    if cached_platform is not None and cached_platform != sys.platform.lower():
        _logprint(uselog, logging.CRITICAL,
                  'please run detect.py before using this script (platform mismatch)')
        sys.exit()

    found = {}
    for device_type, device_details in cache.items():
        if device_types is not None and device_type not in device_types:
            continue

        serial_config = device_details[0]

        # there may be multiple devices for this device type
        for device in device_details[1]:
            port, manufacturer, model, serial_number, _, channels, release_delay = device
            found[port] = (serial_config, device_type, serial_number, model, manufacturer,
                           channels, release_delay)
            snu = serial_number if serial_number else 'unknown'
            _logprint(uselog, logging.INFO,
                      f'cache: {manufacturer} {model} serial number {snu} on {port}')

    if not found:
        _logprint(uselog, logging.CRITICAL,
                  'cache did not contain any matching devices')
        sys.exit()

    return found


def ports_to_channels(settings, found):
    """
    Convert the per-port data structure read from the cache file (as written
    by detect.py), and convert it to a per-channel data structure that is
    necessary for scripts that need to manage multiple-channel power supplies.

    Ignores anything in 'found' that isn't a power supply.

    --------------------------------------------------------------------------
    args
        settings : dict
            contains core information about the test environment
        found : dict
            {port: ({port_config}, device_type,
                    device_serial_number, model, channels), ...}
            e.g.
            {'/dev/cu.usbserial-AH06DY15': ({'baudrate': 9600, 'bytesize': 8,
                                             'parity': 'N', 'stopbits': 1,
                                             'xonxoff': False, 'dsrdtr': False,
                                             'rtscts': False, 'timeout': 1,
                                             'write_timeout': 1,
                                             'inter_byte_timeout': None},
                                            'hvpsu',
                                            '1272738',
                                            '2614b',
                                            'keithley',
                                            ['a', 'b'])}
    --------------------------------------------------------------------------
    returns
        channels : list containing instances of class Channel
            e.g.
            [Channel(port="/dev/cu.usbserial-AH06DY15",
                     config={'baudrate': 9600, 'bytesize': 8, 'parity': 'N',
                             'stopbits': 1, 'xonxoff': False, 'dsrdtr': False,
                             'rtscts': False, 'timeout': 1,
                             'write_timeout': 1, 'inter_byte_timeout': None},
                     serial_number="4428182", model="2614b",
                     manufacturer="keithley", channel="a", category="hvpsu"),
             Channel(port="/dev/cu.usbserial-AH06DY15",
                     config={'baudrate': 9600, 'bytesize': 8, 'parity': 'N',
                             'stopbits': 1, 'xonxoff': False, 'dsrdtr': False,
                             'rtscts': False, 'timeout': 1,
                             'write_timeout': 1, 'inter_byte_timeout': None},
                     serial_number="4428182", model="2614b",
                     manufacturer="keithley", channel="b", category="hvpsu")]
    --------------------------------------------------------------------------
    """
    channels = []

    for port, details in found.items():
        (config, device_type, serial_number, model,
         manufacturer, psu_channels, release_delay) = details

        if device_type not in {'hvpsu', 'lvpsu'}:
            continue

        for channel in psu_channels:
            channels.append(Channel(port, config, serial_number,
                                    model, manufacturer, channel,
                                    device_type, release_delay,
                                    settings['alias']))

    return channels


def exclude_channels(settings, channels):
    """
    Exclude power supply channels from testing that the user has
    specified using the --alias command line option. The default is to leave
    a channel off.

    --------------------------------------------------------------------------
    args
        settings : dictionary
            contains core information about the test environment
        channels : list containing instances of class Channel
            e.g.
            [Channel(port="/dev/cu.usbserial-AH06DY15",
                     config={'baudrate': 9600, 'bytesize': 8, 'parity': 'N',
                             'stopbits': 1, 'xonxoff': False, 'dsrdtr': False,
                             'rtscts': False, 'timeout': 1,
                             'write_timeout': 1, 'inter_byte_timeout': None},
                     serial_number="4428182", model="2614b",
                     manufacturer="keithley", channel="a", category="hvpsu"),
             Channel(port="/dev/cu.usbserial-AH06DY15",
                     config={'baudrate': 9600, 'bytesize': 8, 'parity': 'N',
                             'stopbits': 1, 'xonxoff': False, 'dsrdtr': False,
                             'rtscts': False, 'timeout': 1,
                             'write_timeout': 1, 'inter_byte_timeout': None},
                     serial_number="4428182", model="2614b",
                     manufacturer="keithley", channel="b", category="hvpsu")]
    --------------------------------------------------------------------------
    returns
        channels : list containing instances of class Channel
            e.g. see above
    --------------------------------------------------------------------------
    """
    try:
        return [channel
                for channel in channels
                if _lookup_reference(channel, separator='_') not in settings['ignore']]
    except TypeError:
        return channels


##############################################################################
# motion controller
##############################################################################

def stage_speed(controller_type, stage, speed):
    """
    mm4006 does not have a command to return the stage type, so assume that
    the stage is an ims400ccha as this matches the experimental setup.

    --------------------------------------------------------------------------
    args
        controller_type : string
            name of the controller
        stage : string
            name of the stage
        speed : string
            'slow', 'normal' or 'fast'
    --------------------------------------------------------------------------
    returns : tuple (int, int)
        acceleration, max_velocity
    --------------------------------------------------------------------------
    """
    ils150 = {'slow': (5, 2), 'normal': (5, 5), 'fast': (10, 10)}
    ims400 = {'slow': (6, 4), 'normal': (6, 8), 'fast': (6, 12)}
    accvel = {'ils150pp': ils150, 'ims400': ims400}

    if 'mm4006' in controller_type and 'unknown' in stage:
        stage = 'ims400'

    return accvel[stage][speed]


##############################################################################
# serial port interaction
##############################################################################

def rs232_port_is_valid(com):
    """
    Detect if serial port is an FTDI device.

    Compare each string to each variable; a match in any one is sufficient
    for the whole test to pass. The values in com may be None, 'n/a' or the
    text supplied by a detected device.

    The FTDI USB to RS232 adaptor deployed in the test environment is:
    Startech Network Adapter ICUSB2321F
    https://www.startech.com/en-us/cards-adapters/icusb2321f
    https://uk.rs-online.com/web/p/serial-converters-extenders/1238048/

    Oncology Systems Limited (OSL) recommended USB to RS232 adaptor for IBA
    products is also based on an FTDI chipset (untested):
    https://www.delock.de/produkt/61460/merkmale.html?setLanguage=en

    --------------------------------------------------------------------------
    args
        com : stlp.ListPortInfo
            contains human-readable information regarding the serial port
    --------------------------------------------------------------------------
    returns
        success : bool
            True if this is an FTDI device, False otherwise
    --------------------------------------------------------------------------
    """
    # The Digilent Nexys 4 Artix-7 FPGA Trainer Board used as part of the
    # neutron detector setup shares data with Sam's Windows DAQ application
    # via a shared UART/JTAG USB port. This uses an onboard FTDI device, so
    # ensure no scripts in this suite will interact with it.
    if com.hwid == 'USB VID:PID=0403:6010 SER=210274552605B':
        return False

    success = False
    test_strings = ('FTDI', 'FT232R', 'FT232R')
    test_variables = (com.manufacturer, com.product, com.description)

    for tstr, tvar in zip(test_strings, test_variables):
        try:
            test = tstr in tvar
        except TypeError:
            pass
        else:
            success = success or test
            if test:
                break

    return success


def atomic_send_command_read_response(pipeline, ser, dev, command):
    """
    Atomic send serial port command and receive reply.

    Each serial port associated with a power supply has a lock, and power
    supplies are managed on a per-channel basis. The locks are used to ensure
    that threads controlling channels on multiple-channel PSUs do not attempt
    to use their shared serial port at the same time.

    An exception on the initial serial port write should not happen unless
    the serial port has been disconnected from the computer.

    --------------------------------------------------------------------------
    args
        pipeline : instance of class Production
            contains all the queues through which the production pipeline
            processes communicate
        ser : serial.Serial
            reference for serial port
        dev : instance of class Channel
            contains details of a device and its serial port
        command : bytes
            command to be sent via RS232 to the power supply
    --------------------------------------------------------------------------
    returns
        response : string or None
            returned command string
    --------------------------------------------------------------------------
    """
    # only proceed if no other thread is using this port
    with pipeline.portaccess[dev.port]:
        # if this is a multiple-channel psu, allow some settling time before
        # accessing the serial port
        with contextlib.suppress(TypeError):
            time.sleep(dev.release_delay)

        try:
            ser.write(command)
        except serial.SerialException:
            response = None
            message = f'{dev.ident}, cannot access serial port'
            log_with_colour(logging.ERROR, message)
        else:
            # attempt to read back response
            try:
                response = ser.readline()
            except serial.SerialException:
                response = None
                message = f'{dev.ident}, cannot access serial port'
                log_with_colour(logging.ERROR, message)
            else:
                if dev.manufacturer == 'iseg':
                    # ISEG SHQ replies with local echo on first line
                    # and response on the second line
                    response = ser.readline()

    # AttributeError: handle case where response is None
    with contextlib.suppress(AttributeError):
        response = response.strip().decode('utf-8', errors='replace')

    return response


def send_command(pipeline, ser, dev, command):
    """
    Send serial port command only.

    This is typically used where a response is not expected from the power
    supply and attempting to read a response will only result in a "long" wait
    for the serial port to time out. In general do not use this function with
    ISEG power supplies, since they always echo something, and whatever that
    is needs to be consumed to prevent problems later.

    Each serial port associated with a power supply has a lock, and power
    supplies are managed on a per-channel basis. The locks are used to ensure
    that threads controlling channels on multiple-channel PSUs do not attempt
    to use their shared serial port at the same time.

    An exception on the initial serial port write should not happen unless
    the serial port has been disconnected from the computer.

    --------------------------------------------------------------------------
    args
        pipeline : instance of class Production
            contains all the queues through which the production pipeline
            processes communicate
        ser : serial.Serial
            reference for serial port
        dev : instance of class Channel
            contains details of a device and its serial port
        command : bytes
            command to be sent via RS232 to the power supply
    --------------------------------------------------------------------------
    returns : none
    --------------------------------------------------------------------------
    """
    with pipeline.portaccess[dev.port]:
        # if this is a multiple-channel psu, allow some settling time before
        # accessing the serial port
        with contextlib.suppress(TypeError):
            time.sleep(dev.release_delay)

        try:
            ser.write(command)
        except serial.SerialException:
            message = f'{dev.ident}, cannot access serial port'
            log_with_colour(logging.ERROR, message)


def check_ports_accessible(found, channels, close_after_check=True):
    """
    Check that all the ports listed in cache entries can be successfully
    opened.

    This attempts to open unique serial ports. Hence, for multiple-channel
    devices, only a single attempt is made to open its port.

    --------------------------------------------------------------------------
    args
        found : dict
            {port: ({port_config}, device_type,
                    device_serial_number, model), ...}
            e.g.
            {'/dev/cu.usbserial-AH06DY15': ({'baudrate': 9600, 'bytesize': 8,
                                             'parity': 'N', 'stopbits': 1,
                                             'xonxoff': False, 'dsrdtr': False,
                                             'rtscts': False, 'timeout': 1,
                                             'write_timeout': 1,
                                             'inter_byte_timeout': None},
                                            'hvpsu',
                                            '1272738',
                                            '2410')}
        channels : list containing instances of class Channel
        close_after_check : bool
            defaults to True, psustat.py sets this to False to keep checked
            ports open.
    --------------------------------------------------------------------------
    returns
        spd : dict
            {port: serial_port_identifier, ...}
            E.g.
            {'/dev/ttyUSB0': Serial<id=0xaf4e9c70, open=True>
                             (port='/dev/ttyUSB0', baudrate=9600, bytesize=8,
                              parity='N', stopbits=1, timeout=1,
                              xonxoff=False, rtscts=True, dsrdtr=False),
             '/dev/ttyUSB1': Serial<id=0xaf553b50, open=True>
                             (port='/dev/ttyUSB1', baudrate=9600, bytesize=8,
                              parity='N', stopbits=1, timeout=1,
                              xonxoff=False, rtscts=False, dsrdtr=False)}
            This is ignored by most callers, it's only currently used by
            psustat.py.
        channels : list containing instances of class Channel
            no explicit return - mutable type amended in place
    --------------------------------------------------------------------------
    """
    port_missing = False

    spd = {}
    for port, value in found.items():
        config, *_ = value

        ser = serial.Serial()
        ser.apply_settings(config)
        ser.port = port
        try:
            ser.open()
        except (FileNotFoundError, OSError, serial.SerialException):
            port_missing = True
            message = f'could not open port {port}'
            log_with_colour(logging.ERROR, message)
            break
        else:
            if close_after_check:
                ser.close()
            else:
                spd[port] = ser

    if port_missing:
        # remove everything from channels to prevent any data being taken
        # given that it's probably better to have the user fix the problem
        # with the test configuration, rather than continuing the test
        # with part(s) of the setup inactive
        channels.clear()
        message = 'not all serial ports listed in the cache were accessible'
        log_with_colour(logging.ERROR, message)
        message = 'check connections and/or run ./detect.py'
        log_with_colour(logging.ERROR, message)

    return spd


def missing_ports(found):
    """
    Check that all the ports listed in cache entries can be successfully
    opened.

    --------------------------------------------------------------------------
    args
        found : dict
            {port: ({port_config}, device_type,
                    device_serial_number, model), ...}
            e.g.
            {'/dev/cu.usbserial-AH06DY15': ({'baudrate': 9600, 'bytesize': 8,
                                             'parity': 'N', 'stopbits': 1,
                                             'xonxoff': False, 'dsrdtr': False,
                                             'rtscts': False, 'timeout': 1,
                                             'write_timeout': 1,
                                             'inter_byte_timeout': None},
                                            'hvpsu',
                                            '1272738',
                                            '2410')}
    --------------------------------------------------------------------------
    returns
        port_missing : bool
            True if at least one port could not be successfully opened,
            False otherwise
    --------------------------------------------------------------------------
    """
    port_missing = False

    for port, value in found.items():
        config, *_ = value

        ser = serial.Serial()
        ser.apply_settings(config)
        ser.port = port
        try:
            ser.open()
        except (FileNotFoundError, serial.SerialException):
            port_missing = True
            print(f'could not open port {port}')
        else:
            ser.close()

    return port_missing


##############################################################################
# power supply interaction
##############################################################################

def _channel_mismatch(cache, user):
    """
    Check for a match between the channel identifier from the cache and the
    one supplied by the user.

    Users are likely to incorrectly specify channel identifiers, since some
    manufacturers use letters and some use numbers. If there is an
    alphanumeric mismatch between cache and user values, this alone should not
    be sufficient to remove the channel.

    --------------------------------------------------------------------------
    args
        cache : string, single character
            channel identifier read from cache, e.g. '1', '2', 'a' or 'b'
        user : string, single character
            channel identifier from command line, e.g. '1', '2', 'a' or 'b'
    --------------------------------------------------------------------------
    returns : bool
        True if the two arguments match, False otherwise
    --------------------------------------------------------------------------
    """
    cache_alpha = cache.isalpha()
    if cache_alpha != user.isalpha():
        if cache_alpha:
            user = chr(ord(user) - ord('1') + ord('a'))
        else:
            user = str(ord(user) - ord('a') + 1)

    return cache != user


def initial_power_supply_check(settings, pipeline, psus, channels, psuset=False):
    """
    Establishes RS232 communications with power supplies (as required).
    Checks status of channel outputs and interlocks (inhibits).

    --------------------------------------------------------------------------
    args
        settings : dictionary
            contains core information about the test environment
        pipeline : instance of class Production
            contains all the queues through which the production pipeline
            processes communicate
        psus : dict
            {port: ({port_config}, device_type, device_serial_number), ...}
            contents of the cache filtered by hvpsu category
        channels : list
            contains instances of class Channel, one for each
            power supply channel
        psuset : bool
            selects the error message depending on the caller
    --------------------------------------------------------------------------
    returns
        channels : list
            no explicit return, mutable type amended in place
    --------------------------------------------------------------------------
    """
    port_used = collections.defaultdict(int)
    outstat = []
    intstat = []
    polstat = []

    if settings['debug'] is None:
        check_ports_accessible(psus, channels)

        for dev in channels:
            message = f'enabled: {_lookup_decorative(settings["alias"], dev)}'
            log_with_colour(logging.INFO, message)

            with serial.Serial(port=dev.port) as ser:
                ser.apply_settings(dev.config)

                # try to ensure consistent state
                # clear FTDI output buffer state before sending
                ser.reset_output_buffer()
                # clear PSU state
                if dev.model == '2614b':
                    command_string = lexicon.power(dev.model, 'terminator only',
                                                   channel=dev.channel)
                    send_command(pipeline, ser, dev, command_string)
                # clear FTDI input buffer state
                ser.reset_input_buffer()
                # arbitrary settle time before proceeding
                time.sleep(0.5)

                # ensure serial port communication with PSU
                # this is for ISEG SHQ only, on a per-PSU basis
                if not port_used[dev.port]:
                    synchronise_psu(ser, pipeline, dev)

                # check power supply output
                # but don't do this if called from psuset.py, since the user
                # of that script may be issuing a command to turn a
                # power supply channel output on
                if not psuset:
                    outstat.append(report_output_status(ser, pipeline, dev))

                # check interlock
                # this is per-PSU on Keithley, per-channel on ISEG
                if ((not port_used[dev.port] or dev.manufacturer == 'iseg')
                        and dev.manufacturer not in {'agilent', 'hameg'}):
                    intstat.append(_report_interlock_status(ser, pipeline, dev))

                if dev.manufacturer == 'iseg':
                    polstat.append(_report_polarity_status(settings, ser, pipeline, dev, psuset))

                ser.reset_input_buffer()
                ser.reset_output_buffer()

            port_used[dev.port] += 1

        if any(outstat):
            message = 'all power supply outputs must be on to proceed'
            log_with_colour(logging.ERROR, message)
            channels.clear()

        if any(intstat):
            message = 'all power supply interlocks must be inactive to proceed'
            log_with_colour(logging.ERROR, message)
            channels.clear()

        if any(polstat):
            if psuset:
                message = 'set voltage should agree with polarity switch to proceed'
                log_with_colour(logging.ERROR, message)
            else:
                message = 'all polarity switches must agree with --forwardbias to proceed'
                log_with_colour(logging.ERROR, message)

            channels.clear()
    else:
        message = '--debug: any connected power supplies will be ignored'
        log_with_colour(logging.WARNING, message)
        message = '--debug: IV data will generated internally'
        log_with_colour(logging.WARNING, message)


def rate_limit(timestamp, duration):
    """
    Sleep the thread to limit the rate of change of voltage to the
    device under test. The execution time of the thread is taken into account
    to ensure that the rate of change of voltage is not limited more than
    necessary.

    --------------------------------------------------------------------------
    args
        timestamp : float
            value previously returned by time.time()
        duration : float
            nominal delay time in seconds required to not exceed the desired
            volts-per-second rate limit
    --------------------------------------------------------------------------
    returns : float
        value returned by time.time(), this value will be returned to this
        function at the next call as argument timestamp
    --------------------------------------------------------------------------
    """
    if timestamp is not None:
        tdiff = time.time() - timestamp
        if tdiff < duration:
            time.sleep(duration - tdiff)
    else:
        time.sleep(duration)

    return time.time()


def read_psu_measured_vi(pipeline, ser, dev):
    """
    Read the voltage and current as measured at the psu output terminals
    from a high voltage power supply.

    --------------------------------------------------------------------------
    args
        pipeline : instance of class Production
            contains all the queues through which the production pipeline
            processes communicate
        ser : serial.Serial
            reference for serial port
        dev : instance of class Channel
            contains details of a device and its serial port
    --------------------------------------------------------------------------
    returns
        measured_voltage : float or None
        measured_current : float or None
    --------------------------------------------------------------------------
    """
    measured_voltage = measured_current = None

    if dev.manufacturer in {'agilent', 'iseg'}:
        convert = iseg_value_to_float if dev.manufacturer == 'iseg' else float

        command_string = lexicon.power(dev.model, 'read voltage', channel=dev.channel)
        local_buffer = atomic_send_command_read_response(pipeline, ser, dev, command_string)
        with contextlib.suppress(ValueError):
            measured_voltage = convert(local_buffer)

        command_string = lexicon.power(dev.model, 'read current', channel=dev.channel)
        local_buffer = atomic_send_command_read_response(pipeline, ser, dev, command_string)
        with contextlib.suppress(ValueError):
            measured_current = convert(local_buffer)

        if measured_voltage is None or measured_current is None:
            measured_voltage = measured_current = None

    elif dev.manufacturer == 'keithley':
        command_string = lexicon.power(dev.model, 'read measured vi', channel=dev.channel)
        local_buffer = atomic_send_command_read_response(pipeline, ser, dev, command_string)

        if local_buffer is not None:
            separator = ',' if dev.model == '2410' else None
            items = (float(x) for x in local_buffer.split(separator))

            try:
                measured_voltage = next(items)
                measured_current = next(items)
            except (StopIteration, ValueError):
                measured_voltage = measured_current = None
                message = f'{dev.ident} problem reading measured vi'
                log_with_colour(logging.WARNING, message)

    return measured_voltage, measured_current


def _report_interlock_status(ser, pipeline, dev):
    """
    Check status of power supply interlock.

    This is per-PSU on Keithley, per-channel on ISEG.

    --------------------------------------------------------------------------
    args
        ser : serial.Serial
            reference for serial port
        pipeline : instance of class Production
            contains all the queues through which the production pipeline
            processes communicate
        dev : instance of class Channel
            contains details of a device and its serial port
    --------------------------------------------------------------------------
    returns : none
    --------------------------------------------------------------------------
    """
    assert dev.manufacturer not in {'agilent', 'hameg'},\
        'function not callable for Agilent or Hameg PSU'

    fail = False
    interlock_set = False

    command_string = lexicon.power(dev.model, 'check interlock', channel=dev.channel)
    register = atomic_send_command_read_response(pipeline, ser, dev, command_string)

    if dev.manufacturer == 'iseg':
        # Interlocks are per-channel for ISEG SHQ, so include the channel
        # identifier in the ident string. For consistency with Keithley below
        # do not report the alias to the user.
        ident = _lookup_reference(dev)
        interlock_set = '=INH' in register
    else:
        # Interlocks are per-PSU for Keithley, so omit the channel identifier
        # if it exists.
        ident = _lookup_reference(dev, use_channel=False)
        try:
            regval = int(float(register))
        except ValueError:
            message = f'{ident}, problem checking interlock'
            log_with_colour(logging.WARNING, message)
            fail = True
        else:
            if dev.model == '2614b':
                # bit 11 (status.measurement.INTERLOCK) p.11-280 (648)
                # Without hardware interlock: '0.00000e+00'
                # with hardware interlock: '2.04800e+03'
                interlock_set = regval & 2048 == 0
            elif dev.model == '2410':
                if regval in {0, 1}:
                    # p.18-9 (389)
                    # returns '0' (disabled - no restrictions) or '1' (enabled)
                    interlock_set = register == 0
                else:
                    message = f'{ident}, problem checking interlock'
                    log_with_colour(logging.WARNING, message)

    if interlock_set:
        message = f'{ident}, interlock active'
        log_with_colour(logging.WARNING, message)
        fail = True

    return fail


def report_output_status(ser, pipeline, dev):
    """
    Check that the output for the given power supply channel is on.

    While the outputs may be switched on/off over RS232 for all the supported
    power supplies except ISEG SHQ, this is generally used as a check to make
    sure the user has configured the test environment correctly.

    Values returned in variable output:

    +------------------+-----------+---------------+---------------+
    | dev.manufacturer | dev.model | output OFF    | output ON     |
    +------------------+-----------+---------------+---------------+
    | 'agilent'        | 'e3634a'  | '0'           | '1'           |
    | 'agilent'        | 'e3647a'  | '0'           | '1'           |
    | 'hameg'          | 'hmp4040' | '0'           | '1'           |
    | 'iseg'           | 'shq'     | '=OFF'        | '=ON'         |
    | 'keithley'       | '2410'    | '0'           | '1'           |
    | 'keithley'       | '2614b'   | '0.00000e+00' | '1.00000e+00' |
    +------------------+-----------+---------------+---------------+

    --------------------------------------------------------------------------
    args
        ser : serial.Serial
            reference for serial port
        pipeline : instance of class Production
            contains all the queues through which the production pipeline
            processes communicate
        dev : instance of class Channel
            contains details of a device and its serial port
    --------------------------------------------------------------------------
    returns
        fail : bool
            True if the output was found to be off, False otherwise
    --------------------------------------------------------------------------
    """
    fail = False

    command_string = lexicon.power(dev.model, 'check output', channel=dev.channel)
    output = atomic_send_command_read_response(pipeline, ser, dev, command_string)

    if dev.manufacturer == 'iseg':
        if '=ON' not in output:
            log_with_colour(logging.WARNING, f'{dev.ident}, output off')
            fail = True

    elif dev.manufacturer in {'agilent', 'hameg', 'keithley'}:
        try:
            outval = int(float(output))
        except ValueError:
            message = f'{dev.ident}, problem checking output'
            log_with_colour(logging.WARNING, message)
            fail = True
        else:
            if outval in {0, 1}:
                if outval == 0:
                    message = f'{dev.ident}, output off'
                    log_with_colour(logging.WARNING, message)
                    fail = True
            else:
                message = f'{dev.ident}, problem checking output'
                log_with_colour(logging.WARNING, message)
                fail = True

    return fail


def _report_polarity_status(settings, ser, pipeline, dev, psuset):
    """
    Returned string from 'check module status' command appears to be base 10.

    e.g.
        inhibit is bit 5 (0=inactive, 1=active),
        polarity is bit 2 (0=negative, 1=positive)

        For channel 1, with inhibit active (terminator removed from front
        panel inhibit socket) and the polarity switch set to POS on the
        rear panel, the 'S1' command returns '036'

    e.g.
        polarity set to positive: '004', negative: '000'

    --------------------------------------------------------------------------
    args
        settings : dictionary
            contains core information about the test environment
        ser : serial.Serial
            reference for serial port
        pipeline : instance of class Production
            contains all the queues through which the production pipeline
            processes communicate
        dev : instance of class Channel
            contains details of a device and its serial port
        psuset : bool
            selects the error message depending on the caller
    --------------------------------------------------------------------------
    returns
        fail : bool
            True if the channel's hardware-set polarity was found to conflict
            with the user's forward/reverse bias command line setting,
            False otherwise
    --------------------------------------------------------------------------
    """
    assert dev.model == 'shq', 'function only callable for ISEG SHQ PSU'

    fail = False
    message = ''

    command_string = lexicon.power(dev.model, 'check module status', channel=dev.channel)
    register = atomic_send_command_read_response(pipeline, ser, dev, command_string)

    try:
        polarity_negative = int(register) & 4 == 0
    except ValueError:
        fail = True
        message = f'{dev.ident} problem checking polarity'
        log_with_colour(logging.WARNING, message)
    else:
        poltxt = 'negative' if polarity_negative else 'positive'
        polvol = 'negative' if settings['voltage'] < 0 else 'positive'

        if psuset:
            if settings['voltage'] > 0 != polarity_negative:
                fail = True
                message = (f'{dev.ident}, conflict between set voltage ({polvol}) '
                           f'and rear panel polarity switch ({poltxt})')

        elif settings['forwardbias'] == polarity_negative:
            # forwardbias is a setting from iv.py
            fail = True
            message = (f'{dev.ident}, forward bias setting ({settings["forwardbias"]}) '
                       f'conflicts with rear panel channel polarity switch ({poltxt})')

        if fail:
            log_with_colour(logging.WARNING, message)

    return fail


def set_psu_voltage(settings, pipeline, voltage, ser, dev):
    """
    Set voltage on PSU.

    For the Keithley 2410's 1kV range, voltages can only be specified to two
    decimal places, though setting digits in the second decimal place is
    unreliable (read back values do not always match) so values should be
    limited to one decimal place.

    The ISEG SHQ ramps voltage to the desired value instead of setting it
    instantaneously, therefore poll status to check it has reached the
    desired.

    --------------------------------------------------------------------------
    args
        settings : dictionary
            contains core information about the test environment
        pipeline : instance of class Production
            contains all the queues through which the production pipeline
            processes communicate
        voltage : int
        ser : serial.Serial
            reference for serial port
        dev : instance of class Channel
            contains details of a device and its serial port
    --------------------------------------------------------------------------
    returns : none
    --------------------------------------------------------------------------
    """
    target_voltage = voltage
    if dev.manufacturer == 'iseg':
        # ISEG sets output channel polarity with hardware screws on the
        # rear panel, all voltages issued to the PSU have to be positive.
        voltage = abs(voltage)

    # constrain the voltage to be set to the given number of decimal places
    voltage = decimal_quantize(voltage, settings['decimal_places'])

    # set voltage
    command_string = lexicon.power(dev.model, 'set voltage', voltage, channel=dev.channel)
    if dev.manufacturer in {'agilent', 'keithley', 'hameg'}:
        send_command(pipeline, ser, dev, command_string)
    elif dev.manufacturer == 'iseg':
        atomic_send_command_read_response(pipeline, ser, dev, command_string)
        wait_for_voltage_to_stabilise(ser, pipeline, dev, target_voltage)

    if dev.manufacturer == 'agilent':
        command_string = lexicon.power(dev.model, 'clear event registers', channel=dev.channel)
        atomic_send_command_read_response(pipeline, ser, dev, command_string)


def set_psu_voltage_and_read(settings, pipeline, voltage, ser, dev, settling_time=2):
    """
    Set voltage on PSU then read back the measured voltage and current.

    For the Keithley 2410's 1kV range, voltages can only be specified to two
    decimal places, though setting digits in the second decimal place is
    unreliable (read back values do not always match) so values should be
    limited to one decimal place.

    The ISEG SHQ ramps voltage to the desired value instead of setting it
    instantaneously, therefore poll status to check it has reached the
    desired.

    --------------------------------------------------------------------------
    args
        settings : dictionary
            contains core information about the test environment
        pipeline : instance of class Production
            contains all the queues through which the production pipeline
            processes communicate
        voltage : int
        ser : serial.Serial
            reference for serial port
        dev : instance of class Channel
            contains details of a device and its serial port
        settling_time : int
            seconds
    --------------------------------------------------------------------------
    returns float, float
        measured_voltage, measured_current
    --------------------------------------------------------------------------
    """
    set_psu_voltage(settings, pipeline, voltage, ser, dev)

    # ensure minimum settling time between set and read operations
    time.sleep(settling_time)

    return read_psu_measured_vi(pipeline, ser, dev)


def synchronise_psu(ser, pipeline, dev):
    """
    Only executed for ISEG SHQ power supplies.

    "In order to assure synchronisation between the computer and the supply,
    <CR><LF> has to be sent as first command."

    RS-232 Interface Programmers Guide for SHQ Devices, p.3

    --------------------------------------------------------------------------
    args
        ser : serial.Serial
            reference for serial port
        pipeline : instance of class Production
            contains all the queues through which the production pipeline
            processes communicate
        dev : instance of class Channel
            contains details of a device and its serial port
    --------------------------------------------------------------------------
    returns : none
    --------------------------------------------------------------------------
    """
    if dev.manufacturer == 'iseg':
        command_string = lexicon.power(dev.model, 'synchronise')
        atomic_send_command_read_response(pipeline, ser, dev, command_string)


def unique(settings, psus, channels):
    """
    Remove all power supply channels that do not match the command line
    options that specifically identify which power supply channel to set:

    --manufacturer
    --model
    --serial
    --channel
    --port

    Two measures are used to limit the amount of typing the user has to
    perform:

    (1) for --serial and --port: allow the user to enter
    partial information, e.g. where there are two power supplies:

    cache: keithley 2410 serial number 4343654 on /dev/cu.usbserial-AH06DY15
    cache: keithley 2614b serial number 4428182 on /dev/cu.usbserial-AH06DWHE

    using --serial 654 would be sufficient to identify the keithley 2410
    since 654 does not appear in the other serial number.

    (2) for --channel, allow the user to mix up alphabetic and numeric
    identifiers, so the following pairings are viewed as equivalent:

    '1' and 'a'
    '2' and 'b'

    --------------------------------------------------------------------------
    args
        settings : dict
            contains core information about the test environment
        psus : dict
            {port: ({port_config}, device_type, device_serial_number), ...}
            contents of the cache filtered by hvpsu and lvpsu categories
        channels : list
            contains instances of class Channel, one for each
            power supply channel
    --------------------------------------------------------------------------
    returns
        single : bool
            True if only a single channel is left, False otherwise
        psus : dict
            no explicit return, mutable type amended in place
        channels : list
            no explicit return, mutable type amended in place
    --------------------------------------------------------------------------
    """
    # user supplied command line values
    user = (settings['manufacturer'], settings['model'], settings['serial'],
            settings['port'], settings['channel'])

    # tests for manufacturer, model, serial_number and port
    def generic(cached_values, user_value):
        return user_value not in cached_values

    test = (generic, generic, generic, generic, _channel_mismatch)

    for channel in copy.deepcopy(channels):
        cache = (channel.manufacturer, channel.model, channel.serial_number,
                 channel.port, channel.channel)
        if any(t(c, u) for c, u, t in zip(cache, user, test) if u is not None):
            channels.remove(channel)

    single = len(channels) == 1
    if single:
        # discard surplus entries in psus data structure: only retain power
        # supplies with ports matching those remaining in channels
        channel_ports = {channel.port for channel in channels}
        for psu_port in set(psus):
            if psu_port not in channel_ports:
                del psus[psu_port]

    return single


def wait_for_voltage_to_stabilise(ser, pipeline, dev, target_voltage):
    """
    Because ISEG SHQ ramps up voltage instead of setting it instantaneously,
    need to poll status to check it has reached its destination.

    Note that even when the response to the command contains ON - indicating
    that the power supply deems that the voltage set has been reached - the
    front panel display and the measured voltage both indicate that there is
    still some additional settling time.

    The front panel display has less resolution available than the value read
    back over the serial port connection, and does show some rounding errors.

    FIXME - note that the ISEG SHQ often doesn't want to settle to the set
    voltage value, particularly in the -5 < n < 5 range. e.g. d1=1, then check
    with d1 '00010-10' to indicate the PSU has understood the command, then
    minutes later u1 still shows it has not converged '-00022-01'. It may be
    better to simply perform a ramp and sample at 1s intervals. The ISEG SHQ
    222M (2x 2kV / 6mA) seems to perform better in this regard than the ISEG
    SHQ 224M (2x 4kV / 3mA)

    --------------------------------------------------------------------------
    args
        ser : serial.Serial
            reference for serial port
        pipeline : instance of class Production
            contains all the queues through which the production pipeline
            processes communicate
        dev : instance of class Channel
            contains details of a device and its serial port
        target_voltage : numeric
            FIXME the sign of this argument may not match the measured voltage
            as read back from the instrument - the caller should make sure it
            matches
    --------------------------------------------------------------------------
    returns : none
    --------------------------------------------------------------------------
    """
    assert dev.model == 'shq', 'function only callable for ISEG SHQ PSU'
    target_voltage = float(target_voltage)

    while True:
        command_string = lexicon.power(dev.model, 'check output', channel=dev.channel)
        local_buffer = atomic_send_command_read_response(pipeline, ser, dev, command_string)

        if '=ON' in local_buffer:
            break

        time.sleep(0.5)

    # the psu voltage may still need to stabilise
    for _ in range(16):
        command_string = lexicon.power(dev.model, 'read voltage', channel=dev.channel)
        local_buffer = atomic_send_command_read_response(pipeline, ser, dev, command_string)
        measured_voltage = iseg_value_to_float(local_buffer)

        diff = abs(abs(target_voltage) - abs(measured_voltage))
        if diff <= 0.5 or diff < abs(target_voltage) * 0.04:
            break

        time.sleep(1)
    else:
        mev = f'{si_prefix(measured_voltage)}V'
        tav = f'{si_prefix(target_voltage)}V'
        message = (f'{dev.ident}, measured voltage {mev} '
                   f'did not converge to set voltage {tav}')
        log_with_colour(logging.WARNING, message)


##############################################################################
# logging
##############################################################################

def log_with_colour(level, message):
    """
    Write messages to the log file. This can be safely called from threads,
    but NOT from processes.

    "Thread Safety

    The logging module is intended to be thread-safe without any special work
    needing to be done by its clients. It achieves this though using threading
    locks; there is one lock to serialize access to the module’s shared data,
    and each handler also creates a lock to serialize access to its underlying
    I/O."

    https://docs.python.org/3/library/logging.html#thread-safety

    --------------------------------------------------------------------------
    args
        level : int
            logging level e.g. logging.DEBUG
        message : string
            message to be sent to the log file
    --------------------------------------------------------------------------
    returns : none
    --------------------------------------------------------------------------
    """
    if level == logging.WARNING:
        message = (f'{ANSIColours.FG_BLACK}{ANSIColours.BG_YELLOW}'
                   f'{ANSIColours.BOLD}{message}{ANSIColours.ENDC}')
    elif level in {logging.ERROR, logging.CRITICAL}:
        message = (f'{ANSIColours.FG_WHITE}{ANSIColours.BG_RED}'
                   f'{ANSIColours.BOLD}{message}{ANSIColours.ENDC}')

    logging.log(level, message)



#!/usr/bin/env python3
"""
Set the voltage on a power supply connected by RS232.

All power supplies supported by detect.py are usable by this script.

the script will return an appropriate unix return value on completion, so
the return value can be used by the shell, e.g.

./psuset.py -50 && ./capture2.py -t 60
"""

import argparse
import logging
import math
import sys
import threading
import time

import serial
import numpy as np

from mmcb import common
from mmcb import lexicon
from mmcb import sequence


##############################################################################
# command line option handler
##############################################################################

def check_voltage(val):
    """
    check voltage range

    --------------------------------------------------------------------------
    args
        val : float
            bias voltage
    --------------------------------------------------------------------------
    returns : float
    --------------------------------------------------------------------------
    """
    val = float(val)
    if not -1100 <= val <= 1100:
        raise argparse.ArgumentTypeError(
            f'{val}: '
            'voltage value should be between -1100 and 1100')
    return val


def check_arguments(settings):
    """
    handle command line options

    --------------------------------------------------------------------------
    args
        settings : dictionary
            contains core information about the test environment
    --------------------------------------------------------------------------
    returns : none
    --------------------------------------------------------------------------
    """
    parser = argparse.ArgumentParser(
        description='Set voltage and/or current limit of a single power\
        supply channel connected by RS232. Currently supported PSUs are: Hameg\
        (Rohde & Schwarz) HMP4040; Keithley 2410, 2614b; ISEG SHQ 222M, 224M;\
        and Agilent E3647A, E3634A (the latter may also be branded Keysight\
        or Hewlett-Packard). When multiple power supplies are\
        connected, use command line options --manufacturer, --model,\
        --serial, --channel and --port to identify the individual power\
        supply. For brevity, specify the minimum number of identifying\
        parameters to uniquely identify the device; if there is only one\
        single-channel power supply attached, just specify the voltage to be\
        set. Note that ISEG power supplies can take a long time to settle to\
        final values, and at low voltages they may never converge to the set\
        voltage.')
    parser.add_argument(
        'voltage', nargs='*', metavar='voltage',
        help='value in volts, values range -1100 to +1100',
        type=check_voltage, default=None)
    parser.add_argument(
        '--verbose',
        action='store_true',
        help='Display all power supply channels detected, and search terms\
        supplied')
    parser.add_argument(
        '--manufacturer', nargs=1, metavar='manufacturer',
        choices=['agilent', 'hameg', 'iseg', 'keithley'],
        help='PSU manufacturer.',
        default=None)
    parser.add_argument(
        '--model', nargs=1, metavar='model',
        choices=['2410', '2614b', 'e3634a', 'e3647a', 'hmp4040', 'shq'],
        help='PSU model.',
        default=None)
    parser.add_argument(
        '--serial', nargs=1, metavar='serial',
        help='PSU serial number. A part of the serial may be supplied if it\
        is unique amongst connected devices.',
        default=None)
    parser.add_argument(
        '--channel', nargs=1, metavar='channel',
        choices=['1', '2', '3', '4', 'a', 'A', 'b', 'B', 'c', 'C', 'd', 'D'],
        help='PSU channel number. These can be specified numerically or\
        alphabetically.',
        default=None)
    parser.add_argument(
        '--port', nargs=1, metavar='port',
        help='Serial port identifier. This is useful where multiple Agilent/\
        Keysight/Hewlett-Packard power supplies are connected, since they do\
        not provide a serial number over RS232. A part of the serial may be\
        supplied if it is unique amongst connected devices.',
        default=None)
    parser.add_argument(
        '--reset',
        action='store_true',
        help='Reset power supply before setting voltage (Keithley 2410).')
    parser.add_argument(
        '-i', '--immediate',
        action='store_true',
        help='On high-voltage power supplies, set the specified voltage\
            immediately. By default the voltage will ramp from the current\
            value to the specified value (Keithley and ISEG). For low-voltage\
            power supplies the specified voltage is always set immediately\
            unless --peltier is used with the HMP4040.')
    parser.add_argument(
        '-l', '--limit', nargs=1, metavar='current_limit',
        help='Set the power supply channel current limit.\
        Values can be specified with either scientific (10e-9) or\
        engineering notation (10n). Supported for Keithley 2410, 2614b;\
        Rohde & Schwarz HMP4040. For the latter, the minimum value is 1mA.',
        type=common.check_current)
    parser.add_argument(
        '-p', '--peltier',
        action='store_true',
        help='Use gradual voltage changes for HMP4040.')
    parser.add_argument(
        '-v', '--voltspersecond', nargs=1, metavar='rateofchange',
        help='When not using --immediate, this option sets the rate of change\
            in volts per second. The default is 10V/s (Keithley and ISEG).')

    group1 = parser.add_mutually_exclusive_group()
    group1.add_argument(
        '-f', '--front',
        action='store_true',
        help='Use front output (Keithley 2410).')
    group1.add_argument(
        '-r', '--rear',
        action='store_true',
        help='Use rear output (Keithley 2410).')

    group2 = parser.add_mutually_exclusive_group()
    group2.add_argument(
        '--on',
        action='store_true',
        help='Turn PSU (channel) output on, by default the script leaves the\
            power supply output unchanged (Keithley and Agilent).')
    group2.add_argument(
        '--off',
        action='store_true',
        help='Turn PSU output off, by default the script leaves the power\
            supply output unchanged (Keithley and Agilent).')

    args = parser.parse_args()

    settings['verbose'] = args.verbose

    # default: leave power supply settings unchanged
    if args.front:
        settings['rear'] = False
    elif args.rear:
        settings['rear'] = True

    # default: leave power supply settings unchanged
    if args.on:
        settings['on'] = True
    elif args.off:
        settings['on'] = False

    if args.reset:
        settings['reset'] = args.reset

    if args.immediate:
        settings['immediate'] = args.immediate

    if args.peltier:
        settings['peltier'] = args.peltier

    # voltage will always be present
    if args.voltage is not None and len(args.voltage) == 1:
        settings['voltage'] = args.voltage[0]

    if args.serial:
        settings['serial'] = args.serial[0]

    if args.channel:
        settings['channel'] = args.channel[0].lower()

    if args.manufacturer:
        settings['manufacturer'] = args.manufacturer[0]

    if args.model:
        settings['model'] = args.model[0]

    if args.port:
        settings['port'] = args.port[0]

    if args.voltspersecond:
        settings['voltspersecond'] = int(args.voltspersecond[0])

    if args.limit:
        settings['current_limit'] = args.limit[0]


##############################################################################
# set psu values
##############################################################################

def read_measured_voltage(settings, pipeline, ser, dev):
    """
    read the voltage as measured at the psu output terminals

    --------------------------------------------------------------------------
    args
        settings : dictionary
            contains core information about the test environment
        pipeline : instance of class Production
            contains all the queues through which the production pipeline
            processes communicate
        ser : serial.Serial
            reference for serial port
        dev : instance of class Channel
            contains details of a device and its serial port
    --------------------------------------------------------------------------
    returns
        success : bool
    --------------------------------------------------------------------------
    """
    measured_voltage = None

    if dev.manufacturer == 'iseg':
        command_string = lexicon.power(dev.model, 'read voltage', channel=dev.channel)
        local_buffer = common.atomic_send_command_read_response(pipeline, ser, dev,
                                                                command_string)

        measured_voltage = common.iseg_value_to_float(local_buffer)

    elif dev.manufacturer == 'keithley':
        command_string = lexicon.power(dev.model, 'read voltage', channel=dev.channel)
        local_buffer = common.atomic_send_command_read_response(pipeline, ser, dev,
                                                                command_string)

        if local_buffer is not None:
            separator = ',' if dev.model == '2410' else None
            items = (float(x) for x in local_buffer.split(separator))

            try:
                measured_voltage = next(items)
            except (StopIteration, ValueError):
                message = f'{dev.ident} problem reading measured voltage'
                common.log_with_colour(logging.WARNING, message)

    elif dev.manufacturer in {'agilent', 'hameg'}:
        if dev.model == 'e3634a':
            command_string = lexicon.power(dev.model, 'set remote')
            common.send_command(pipeline, ser, dev, command_string)

        command_string = lexicon.power(dev.model, 'read voltage', channel=dev.channel)
        local_buffer = common.atomic_send_command_read_response(pipeline, ser, dev,
                                                                command_string)

        try:
            volt = float(local_buffer)
        except ValueError:
            pass
        else:
            measured_voltage = common.decimal_quantize(volt, settings['decimal_places'])

    return measured_voltage


def check_measured_voltage(settings, pipeline, ser, dev, set_voltage):
    """
    compare the voltage as measured at the psu output terminals to the
    given voltage

    --------------------------------------------------------------------------
    args
        settings : dictionary
            contains core information about the test environment
        pipeline : instance of class Production
            contains all the queues through which the production pipeline
            processes communicate
        ser : serial.Serial
            reference for serial port
        dev : instance of class Channel
            contains details of a device and its serial port
        set_voltage : decimal.Decimal
    --------------------------------------------------------------------------
    returns
        success : bool
    --------------------------------------------------------------------------
    """
    success = False
    measured_voltage = read_measured_voltage(settings, pipeline, ser, dev)

    # comparison
    if measured_voltage is not None:
        set_voltage = common.decimal_quantize(set_voltage, settings['decimal_places'])
        measured_voltage = common.decimal_quantize(measured_voltage, settings['decimal_places'])

        volt = common.si_prefix(measured_voltage)
        common.log_with_colour(logging.INFO, f'measured: {volt}V')

        if set_voltage == measured_voltage:
            success = True

    return success


def current_limit(settings, pipeline, ser, dev):
    """
    Set and check current limit on Hameg HMP4040, Keithley 2614b and 2410.

    HMP4040 notes:

    The PSU seems to support setting of values down to 100uA for current limit
    values < 1A, and down to 1mA for current limit >= 1A. However, support
    seems patchy. 0.9992A, 0.0012 and 1.2m set correctly, but 0.0002A, 0.0002
    and 0.2m do not. So limit this to 1mA (3 decimal places) to avoid
    confusion.

    --------------------------------------------------------------------------
    args
        settings : dictionary
            contains core information about the test environment
        pipeline : instance of class Production
            contains all the queues through which the production pipeline
            processes communicate
        ser : serial.Serial
            reference for serial port
        dev : instance of class Channel
            contains details of a device and its serial port
    --------------------------------------------------------------------------
    returns : none
    --------------------------------------------------------------------------
    """
    if dev.model not in ('hmp4040', '2614b', '2410'):
        return

    # set current limit
    if dev.model == 'hmp4040':
        dec_places = 3
        compliance = f'{settings["current_limit"]:.{dec_places}f}'
    else:
        compliance = f'{settings["current_limit"]:e}'

    command_string = lexicon.power(dev.model, 'set current limit',
                                   compliance, channel=dev.channel)
    common.send_command(pipeline, ser, dev, command_string)

    # read back current limit from PSU
    command_string = lexicon.power(dev.model, 'get current limit', channel=dev.channel)
    response = common.atomic_send_command_read_response(pipeline, ser, dev, command_string)
    one_percent = 0.005
    onehundred_microamps = 0.0001
    close_enough = math.isclose(float(response), settings['current_limit'],
                                rel_tol=one_percent, abs_tol=onehundred_microamps)

    # report current limit
    message = ('requested current limit '
               f'{common.si_prefix(settings["current_limit"])}A, '
               f'set to {common.si_prefix(response)}A')
    if close_enough:
        common.log_with_colour(logging.INFO, message)
    else:
        common.log_with_colour(logging.WARNING, message)


def configure_lvpsu(settings, pipeline, ser, dev):
    """
    Set the power supply voltage and read back the value to confirm it has
    been correctly set.

    Handles: Agilent E3634A, E3647A; Hameg HMP4040.

    For Agilent, though the low range is set by default, if the voltage value
    submitted exceeds the range threshold by over a volt or so, the PSU will
    automatically select the high range.

    --------------------------------------------------------------------------
    args
        settings : dictionary
            contains core information about the test environment
        pipeline : instance of class Production
            contains all the queues through which the production pipeline
            processes communicate
        ser : serial.Serial
            reference for serial port
        dev : instance of class Channel
            contains details of a device and its serial port
    --------------------------------------------------------------------------
    returns
        success : bool
            True if the device was found, False otherwise
    --------------------------------------------------------------------------
    """
    success = False

    if settings['reset']:
        if dev.manufacturer == 'hameg':
            # reset to default conditions, output off
            # allow device time to complete reset before sending more commands
            command_string = lexicon.power(dev.model, 'reset')
            common.send_command(pipeline, ser, dev, command_string)
            common.log_with_colour(logging.INFO, 'reset')
            time.sleep(0.2)
        else:
            message = f'{dev.manufacturer} {dev.model}: reset not supported'
            common.log_with_colour(logging.WARNING, message)

    # set and verify current limit
    if settings['current_limit'] is not None:
        current_limit(settings, pipeline, ser, dev)

    # change output on/off if requested, otherwise do not change it
    if settings['on'] is not None:
        comtxt = 'output on' if settings['on'] else 'output off'
        command_string = lexicon.power(dev.model, comtxt, channel=dev.channel)
        common.atomic_send_command_read_response(pipeline, ser, dev, command_string)
        common.log_with_colour(logging.INFO, comtxt)

    # set voltage
    if settings['voltage'] is not None:
        # constrain the voltage to be set to the given number of decimal places
        voltage = common.decimal_quantize(settings['voltage'], settings['decimal_places'])

        if voltage < 0:
            message = f'{dev.manufacturer} {dev.model}: cannot use a negative voltage'
            common.log_with_colour(logging.ERROR, message)
        else:
            message = f'voltage set to {common.si_prefix(voltage)}V'
            common.log_with_colour(logging.INFO, message)

        if dev.manufacturer == 'hameg' and settings['peltier']:
            measured_voltage = read_measured_voltage(settings, pipeline, ser, dev)
            if measured_voltage is not None:
                transition_voltage(settings, pipeline, measured_voltage, voltage, ser, dev)
        else:
            common.set_psu_voltage(settings, pipeline, voltage, ser, dev)

        # Agilent E3647A seems to require a minimum of half a second to apply
        # the setting before any further RS232 interaction is reliable
        time.sleep(0.5)

        # Arguably the voltage set for the channel should be read back too, so
        # for the case where the channel output is off, some confirmation that
        # the value has been set correctly can be given to the user.
        #
        # check measured voltage (can only be read back if output is on)
        if not common.report_output_status(ser, pipeline, dev):
            # allow a little time for the voltage to settle before reading
            time.sleep(0.1)
            success = check_measured_voltage(settings, pipeline, ser, dev, voltage)
        else:
            message = 'with PSU output switched off, set voltage cannot be read back'
            common.log_with_colour(logging.WARNING, message)
            common.log_with_colour(logging.WARNING, 'assuming all is well')
            success = True

    return success


def configure_hvpsu(settings, pipeline, ser, dev):
    """
    Amend power supply settings. If a change of voltage has been requested,
    read back the value to confirm it has been correctly set.

    Process requests relevant to the selected power supply in this order:

    * reset
    * set range
    * select front/rear output
    * output on/off
    * set voltage

    Handles: Keithley 2410, 2614b; ISEG SHQ.

    --------------------------------------------------------------------------
    args
        settings : dictionary
            contains core information about the test environment
        pipeline : instance of class Production
            contains all the queues through which the production pipeline
            processes communicate
        ser : serial.Serial
            reference for serial port
        dev : instance of class Channel
            contains details of a device and its serial port
    --------------------------------------------------------------------------
    returns
        success : bool
            True if the device was found, False otherwise
    --------------------------------------------------------------------------
    """
    success = False

    if dev.manufacturer == 'keithley':
        command_string = lexicon.power(dev.model, 'clear event registers')
        common.send_command(pipeline, ser, dev, command_string)

    if settings['reset']:
        if dev.manufacturer == 'keithley' and dev.model == '2410':
            # reset to default conditions, output off
            # allow device time to complete reset before sending more commands
            common.log_with_colour(logging.INFO, 'reset')
            command_string = lexicon.power(dev.model, 'reset')
            common.send_command(pipeline, ser, dev, command_string)
            time.sleep(0.2)
        else:
            message = f'{dev.manufacturer} {dev.model}: reset not supported'
            common.log_with_colour(logging.WARNING, message)

    # set and verify current limit
    if settings['current_limit'] is not None:
        current_limit(settings, pipeline, ser, dev)

    # set voltage
    if settings['voltage'] is not None:
        # as a general rule this script should avoid making changes to the
        # user's power supply settings, since the context it's being used in
        # may be unclear. However, for this project the highest voltage range
        # will always be used, and it is not unreasonable to set it here.
        #
        # It would be more correct for 'configure range' to be performed
        # immediately after a reset (above, selectively). Performing it here
        # (always) is more friendly to the user, e.g. if the user has just
        # switched on the power supply (it may well be in a low voltage range)
        # and runs this script with a high voltage, it will generate a series
        # of
        # "settings conflict" errors (on the 2410) as the script tries to set
        # voltages higher than the current range allows.
        common.log_with_colour(logging.INFO, 'configure range')
        command_string = lexicon.power(dev.model, 'configure range', channel=dev.channel)
        common.send_command(pipeline, ser, dev, command_string)

    # Change front/rear routing if requested, otherwise do not change it.
    #
    # Since changing this setting will turn the output off, it must be performed
    # before processing any user request to turn the output on or off.
    if settings['rear'] is not None:
        if dev.manufacturer == 'keithley' and dev.model == '2410':
            destination = 'REAR' if settings['rear'] else 'FRON'
            command_string = lexicon.power(dev.model, 'set route', destination)
            common.send_command(pipeline, ser, dev, command_string)
        else:
            message = f'{dev.manufacturer} {dev.model}: front/rear not supported'
            common.log_with_colour(logging.WARNING, message)

    # change output on/off if requested, otherwise do not change it
    if settings['on'] is not None:
        if dev.manufacturer == 'keithley':
            comtxt = 'output on' if settings['on'] else 'output off'
            common.log_with_colour(logging.INFO, comtxt)
            command_string = lexicon.power(dev.model, comtxt, channel=dev.channel)
            common.atomic_send_command_read_response(pipeline, ser, dev, command_string)
        else:
            message = f'{dev.manufacturer} {dev.model}: output on/off not supported'
            common.log_with_colour(logging.WARNING, message)

    ##############################################################################
    # set voltage
    if settings['voltage'] is not None:
        # constrain the voltage to be set to the given number of decimal places
        voltage = common.decimal_quantize(settings['voltage'], settings['decimal_places'])
        message = f'setting voltage: {common.si_prefix(voltage)}V'
        common.log_with_colour(logging.INFO, message)

        # set voltage
        if settings['immediate']:
            common.set_psu_voltage(settings, pipeline, voltage, ser, dev)
        else:
            measured_voltage = read_measured_voltage(settings, pipeline, ser, dev)
            if measured_voltage is not None:
                transition_voltage(settings, pipeline, measured_voltage, voltage, ser, dev)

        # check set voltage (can only be read back if output is on)
        if not common.report_output_status(ser, pipeline, dev):
            # allow a little time for the voltage to settle before reading
            time.sleep(0.5)
            success = check_measured_voltage(settings, pipeline, ser, dev, voltage)
        else:
            # keithley 2410, 2614b
            message = 'with PSU output switched off, set voltage cannot be read back'
            common.log_with_colour(logging.WARNING, message)
            common.log_with_colour(logging.WARNING, 'assuming all is well')
            success = True

    return success


def transition_voltage(settings, pipeline, initial_voltage, target_voltage, ser, dev):
    """
    Gradual voltage transitions.

    --------------------------------------------------------------------------
    args
        settings : dictionary
            contains core information about the test environment
        pipeline : instance of class Production
            contains all the queues through which the production pipeline
            processes communicate
        initial_voltage : float
            set voltage at present
        target_voltage : float
            voltage to transition to
        ser : serial.Serial
            reference for serial port
        dev : instance of class Channel
            contains details of a device and its serial port
    --------------------------------------------------------------------------
    returns : none
    --------------------------------------------------------------------------
    """
    if initial_voltage != target_voltage:
        message = f'{dev.ident}, transitioning from {initial_voltage}V to {target_voltage}V'
        common.log_with_colour(logging.INFO, message)

        if dev.manufacturer == 'iseg':
            # use power supply's internal ramp feature to avoid having manage
            # the ISEG SHQ's tardy response time

            # read voltage rate of change
            command_string = lexicon.power(dev.model,
                                           'read max rate of change',
                                           channel=dev.channel)
            max_vroc = common.atomic_send_command_read_response(pipeline, ser, dev,
                                                                command_string)

            # set voltage rate of change for ramp
            if settings['voltspersecond'] is None:
                volts_per_second = 10
            else:
                volts_per_second = settings['voltspersecond']

            command_string = lexicon.power(dev.model,
                                           'set voltage max rate of change',
                                           volts_per_second,
                                           channel=dev.channel)
            common.atomic_send_command_read_response(pipeline, ser, dev, command_string)

            # auto ramp will progress towards the given voltage at the given rate
            command_string = lexicon.power(dev.model,
                                           'set voltage',
                                           abs(target_voltage),
                                           channel=dev.channel)

            common.atomic_send_command_read_response(pipeline, ser, dev, command_string)

            # wait to reach given voltage
            common.wait_for_voltage_to_stabilise(ser, pipeline, dev, target_voltage)

            # revert voltage rate of change to original value
            command_string = lexicon.power(dev.model,
                                           'set voltage max rate of change',
                                           max_vroc,
                                           channel=dev.channel)
            common.atomic_send_command_read_response(pipeline, ser, dev, command_string)

        elif dev.manufacturer == 'keithley':
            timestamp = None
            duration = 1 if settings['voltspersecond'] is None else 10 / settings['voltspersecond']
            step = 10

            for voltage in sequence.to_original(initial_voltage, target_voltage, step):
                timestamp = common.rate_limit(timestamp, duration)
                common.set_psu_voltage(settings, pipeline, voltage, ser, dev)

            common.log_with_colour(logging.INFO, f'{dev.ident}, transition complete')

        elif dev.model == 'hmp4040':
            # quick addition - special case for Peltier usage
            # aim for 2V in 30s, 1.5s per 0.1V step
            initial_voltage = float(initial_voltage)
            target_voltage = float(target_voltage)
            timestamp = None
            step = 0.1 if target_voltage > initial_voltage else -0.1
            duration = 30 / (2 / abs(step))

            for voltage in np.arange(initial_voltage + step, target_voltage + step, step):
                timestamp = common.rate_limit(timestamp, duration)
                common.set_psu_voltage(settings, pipeline, voltage, ser, dev)


##############################################################################
# main
##############################################################################

def main():
    """ set the voltage on a power supply connected by RS232 """
    success = False
    settings = {
        'alias': None,
        'channel': None,
        'current_limit': None,
        'debug': None,
        'decimal_places': 3,
        'immediate': False,
        'manufacturer': None,
        'model': None,
        'on': None,
        'peltier': False,
        'port': None,
        'rear': None,
        'reset': False,
        'serial': None,
        'verbose': None,
        'voltage': None,
        'voltspersecond': 10
        }

    check_arguments(settings)

    ##########################################################################
    # enable logging to screen
    ##########################################################################

    logging.basicConfig(
        level=logging.DEBUG,
        format='%(asctime)s : %(levelname)s : %(message)s')

    # enable logging to screen
    console = logging.StreamHandler()
    console.setLevel(logging.INFO)

    # set logging time to UTC to match session timestamp
    logging.Formatter.converter = time.gmtime

    ##########################################################################
    # filter power supplies from cache leaving just a single channel
    ##########################################################################

    psus = common.cache_read(['hvpsu', 'lvpsu'])
    channels = common.ports_to_channels(settings, psus)

    if settings['verbose']:
        print('detected power supply channels:')
        for channel in channels:
            print(channel)
        print('search terms:')
        maxlen = max(len(x) for x in settings)
        for setting, value in settings.items():
            print(f'{setting:>{maxlen}} {value}')

    if not common.unique(settings, psus, channels):
        print('could not identify a single power supply channel')
        print('check use of --manufacturer, --model, --serial, --port, and --channel')
        sys.exit()

    ##########################################################################
    # set up resources for threads
    #
    # this is overkill since only one power supply channel will be used
    # but does allow the same well-tested interface employed by other scripts
    # in the suite to be used
    ##########################################################################

    class Production:
        """ Locks to support threaded operation. """
        portaccess = {port: threading.Lock()
                      for port in {channel.port for channel in channels}}

    pipeline = Production()

    ##########################################################################
    # Check status of outputs and interlock (inhibit) on all power supplies
    ##########################################################################

    common.initial_power_supply_check(settings, pipeline, psus, channels, psuset=True)

    ##########################################################################
    # display details of selected power supply
    ##########################################################################

    try:
        channel = channels[0]
    except IndexError:
        pass
    else:
        text = f'selected power supply: {channel.manufacturer} {channel.model}'
        if channel.serial_number:
            text += f' s.no. {channel.serial_number}'
        if channel.channel:
            text += f' channel {channel.channel}'
        if settings['port'] is not None:
            text += f' port {channel.port}'
        common.log_with_colour(logging.INFO, text)

        ##########################################################################
        # set values on given power supply channel
        ##########################################################################

        setpsu = {'lvpsu': configure_lvpsu, 'hvpsu': configure_hvpsu}

        with serial.Serial(port=channel.port) as ser:
            ser.apply_settings(channel.config)
            ser.reset_input_buffer()
            ser.reset_output_buffer()
            success = setpsu[channel.category](settings, pipeline, ser, channel)
            if success:
                common.log_with_colour(logging.INFO, 'operation successful')

    ##########################################################################
    # 0 command completed successfully
    # 3 non-reserved error code
    return 0 if success else 3


##############################################################################
if __name__ == '__main__':
    sys.exit(main())



#!/usr/bin/env python3
"""
Change power supply polarity supplied to Peltier devices using pairs of SPDT
relays.

Assume we are using a SparkFun Qwiic HAT for Raspberry Pi [1] and one or
more Raspberry Pi Pico Relay Boards [2].

[1] https://www.sparkfun.com/products/14459
[2] https://shop.sb-components.co.uk/products/raspberry-pi-pico-relay-board
"""

import argparse
import sys
import types

from RPi import GPIO

from mmcb.common import ANSIColours


##############################################################################
# data structures
##############################################################################

class Peltier:
    """
    Peltier device.
    """
    __slots__ = {
        'device': 'Numeric identifier for the Peltier device.',
        'board': 'The relay board this device is controlled from.',
        'relay_0': 'The first of the pair of relays controlling this device.',
        'relay_1': 'The second of the pair of relays controlling this device.',
        'gpio': 'Raspberry Pi GPIO pin controlling both relays.'}

    def __init__(self, device, board, relay_0, relay_1, gpio):
        self.device = device
        self.board = board
        self.relay_0 = relay_0
        self.relay_1 = relay_1
        self.gpio = gpio

    def __repr__(self):
        return (f'Peltier('
                f'device={self.device}, '
                f'board={self.board}, '
                f'relay_0={self.relay_0}, '
                f'relay_1={self.relay_1}, '
                f'gpio={self.gpio})')

    def __str__(self):
        on_ansi = f'{ANSIColours.BOLD}{ANSIColours.FG_RED}negative{ANSIColours.ENDC}'
        off_ansi = f'{ANSIColours.BOLD}{ANSIColours.FG_BLUE}positive{ANSIColours.ENDC}'
        relay_stat = on_ansi if self._check_configuration() else off_ansi

        return (f'Peltier {self.device}, board {self.board}, '
                f'relays {self.relay_0} and {self.relay_1} (GPIO {self.gpio:>2}), '
                f'config: {relay_stat}')

    def _check_configuration(self):
        """
        Return a tuple containing the status of both relays controlling this Peltier.
        """
        return GPIO.input(self.gpio)

    def configure_positive(self):
        """
        Set for positive configuration, power supply positive terminal to
        Peltier red wire, power supply negative terminal to Peltier blue
        wire.
        """
        # if already positive, don't do anything
        if self._check_configuration() == GPIO.LOW:
            return True

        GPIO.output(self.gpio, GPIO.LOW)

        return self._check_configuration() == GPIO.LOW

    def configure_negative(self):
        """
        Set for negative configuration, power supply positive terminal to
        Peltier blue wire, power supply negative terminal to Peltier red
        wire.
        """
        # if already negative, don't do anything
        if self._check_configuration() == GPIO.HIGH:
            return True

        GPIO.output(self.gpio, GPIO.HIGH)

        return self._check_configuration() == GPIO.HIGH

    # sufficient to allow a list of class instances to be sorted
    def __lt__(self, value):
        return self.device < value.device


##############################################################################
# command line option handler
##############################################################################

def check_arguments():
    """
    Handle command line options.

    --------------------------------------------------------------------------
    args : none
    --------------------------------------------------------------------------
    returns : class argparse.ArgumentParser
    --------------------------------------------------------------------------
    """
    parser = argparse.ArgumentParser(
        description='Change power supply polarity supplied to a Peltier\
        device using pairs of SPDT relays. With no arguments this script\
        simply reports status of all relays and changes nothing. With just\
        the Peltier device number given, status for that device is given, and\
        again, no changes are made. Note that using just two SPDT\
        relays per Peltier, only configuration of polarity can be set.\
        IMPORTANT: Turning off each Peltier device must be performed using\
        the power supply - the power supply should be turned off before\
        polarity is changed.')
    parser.add_argument(
        'peltier_device_number', nargs='?', metavar='peltier_device_number',
        help='Peltier device identifier.',
        type=int, default=None, choices=[1, 2, 3, 4])

    group1 = parser.add_mutually_exclusive_group()
    group1.add_argument(
        '-p', '--positive',
        action='store_true',
        help='Set positive polarity.')
    group1.add_argument(
        '-n', '--negative',
        action='store_true',
        help='Set negative polarity.')

    return parser.parse_args()


##############################################################################
def main():
    """
    Change power supply polarity supplied to Peltier devices using pairs of
    SPDT relays.
    """
    args = check_arguments()

    return_code = types.SimpleNamespace(
        # Successful operation.
        success=0,
        # Operation not permitted.
        error_eperm=1,
        # I/O error - plausible error code for relays not responding.
        error_eio=5,
        # Invalid - used here for invalid relay configuration
        error_einval=22)

    if not args.peltier_device_number and (args.positive or args.negative):
        print('-p and -n require a peltier device to be specified')
        return return_code.error_eperm

    # each relay board has 4 relays, two per Peltier device
    # {peltier_identifier: [board_identifier, (relay_1, relay_2), gpio], ...}
    peltier_config = {1: [1, (1, 2), 12],  # wire: brown, changed from 21 -> 12 due to faulty GPIO pin 21
                      2: [1, (3, 4), 20],  # wire: violet
                      3: [2, (1, 2), 16],  # wire: black
                      4: [2, (3, 4), 6]}   # wire: yellow

    # configuration of the test setup is incomplete, so only use the relays
    # for devices that are currently in use
    enabled_peltier_devices = {1, 2, 3, 4}

    # if the user has specified an enabled peltier, omit details of other devices
    if args.peltier_device_number:
        # device specified - only print status for the given device if it's enabled
        enabled_peltier_devices = enabled_peltier_devices.intersection({args.peltier_device_number})

    # return promptly with an error code if there are no devices to display
    if not enabled_peltier_devices:
        return return_code.error_eperm

    # basic GPIO configuration
    GPIO.setwarnings(False)
    GPIO.setmode(GPIO.BCM)

    # create data structures for all required Peltier devices
    # and configure GPIO pins as outputs
    peltier_devices = set()
    for device in enabled_peltier_devices:
        board, (relay_0, relay_1), gpio = peltier_config[device]
        GPIO.setup(gpio, GPIO.OUT)
        peltier_devices.add(Peltier(device, board, relay_0, relay_1, gpio))

    # print current status
    status_check_only = not (args.positive or args.negative)
    if status_check_only:
        for device in sorted(peltier_devices):
            print(device)

        return return_code.success

    # at this point, there will only be one device remaining
    device = peltier_devices.pop()

    # process the user's command
    if args.positive:
        success = device.configure_positive()
    elif args.negative:
        success = device.configure_negative()
    else:
        success = False

    return return_code.success if success else return_code.error_einval


##############################################################################
if __name__ == '__main__':
    sys.exit(main())



#!/usr/bin/env python3
"""
Read values from the Keithley DMM6500.
"""

import argparse
import sys
import time

from mmcb import common
from mmcb import dmm_interface


##############################################################################
# command line option handler
##############################################################################


def check_arguments():
    """
    handle command line options

    --------------------------------------------------------------------------
    args : none
    --------------------------------------------------------------------------
    returns
        args : <class 'argparse.Namespace'>
    --------------------------------------------------------------------------
    """
    parser = argparse.ArgumentParser(
        description='Reads voltage or current from the Keithley DMM6500.'
    )
    parser.add_argument(
        '-c', '--capacitance',
        action='store_true',
        help='Read capacitance (F)')
    parser.add_argument(
        '--currentac',
        action='store_true',
        help='Read AC current (A)')
    parser.add_argument(
        '-i', '--currentdc',
        action='store_true',
        help='Read DC current (A)')
    parser.add_argument(
        '-r', '--resistance',
        action='store_true',
        help='Read two-wire resistance (\u03a9)')
    parser.add_argument(
        '-t', '--temperature',
        action='store_true',
        help='Read temperature (\u00b0C)')
    parser.add_argument(
        '--voltageac',
        action='store_true',
        help='Read AC voltage (V)')
    parser.add_argument(
        '-v', '--voltagedc',
        action='store_true',
        help='Read DC voltage (V)')

    args = parser.parse_args()

    return args


##############################################################################
# main
##############################################################################


def main():
    """ Read values from the Keithley DMM6500 """

    args = check_arguments()

    dmm = dmm_interface.Dmm6500()
    if dmm.init_failed:
        print('could not initialise serial port')
        return 3

    configure = {
            'capacitance': dmm.configure_read_capacitance,
            'currentac': dmm.configure_read_ac_current,
            'currentdc': dmm.configure_read_dc_current,
            'resistance': dmm.configure_read_resistance,
            'temperature': dmm.configure_read_temperature,
            'voltagedc': dmm.configure_read_dc_voltage,
            'voltageac': dmm.configure_read_ac_voltage,
    }

    # AC quantities sometimes return None on the first attempt
    retries = 3

    success = []
    for function, required in vars(args).items():
        if not required:
            continue

        configure[function]()

        value = None
        for _ in range(retries):
            value = dmm.read_value()
            if value is not None:
                success.append(True)
                break
            time.sleep(0.1)

        print(f'{function}, {common.si_prefix(value)}, {value}')

    return 0 if any(success) else 3


##############################################################################
if __name__ == '__main__':
    sys.exit(main())
